# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Membership related rules."""

import re
import logging

from contextlib import suppress
from mailman.core.i18n import _
from mailman.interfaces.action import Action
from mailman.interfaces.address import InvalidEmailAddressError
from mailman.interfaces.bans import IBanManager
from mailman.interfaces.member import MemberRole
from mailman.interfaces.rules import IRule
from mailman.interfaces.usermanager import IUserManager
from public import public
from zope.component import getUtility
from zope.interface import implementer


def _find_sender_member(mlist, msg):
    # For every sender email in the message, try to find a member associated
    # with that email.
    #
    # First, check the sender email directly.  If it's a member, we're done.
    #
    # Next, check to see if the sender email is linked to an existing user,
    # and if so, check to see if any of the addresses linked to that user is a
    # member.
    user_manager = getUtility(IUserManager)
    for sender in msg.senders:
        member = mlist.members.get_member(sender)
        if member is not None:
            return member
        user = user_manager.get_user(sender)
        if user is not None:
            for address in user.addresses:
                member = mlist.members.get_member(address.email)
                if member is not None:
                    return member
    return None


@public
@implementer(IRule)
class MemberModeration:
    """The member moderation rule."""

    name = 'member-moderation'
    description = _('Match messages sent by moderated members.')
    record = True

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        # The MemberModeration rule misses unconditionally if any of the
        # senders are banned.
        ban_manager = IBanManager(mlist)
        for sender in msg.senders:
            if ban_manager.is_banned(sender):
                return False
        member = _find_sender_member(mlist, msg)
        if member is None:
            return False
        action = (mlist.default_member_action
                  if member.moderation_action is None
                  else member.moderation_action)
        if action is Action.defer:
            # The regular moderation rules apply.
            return False
        elif action is not None:
            # We must stringify the moderation action so that it can be
            # stored in the pending request table.
            msgdata['member_moderation_action'] = action.name
            msgdata['moderation_sender'] = sender
            with _.defer_translation():
                # This will be translated at the point of use.
                msgdata.setdefault('moderation_reasons', []).append(
                    _('The message comes from a moderated member'))
            return True
        # The sender is not a member so this rule does not match.
        return False


def _do_action(msgdata, action, sender):
    if action is Action.defer:
        # The regular moderation rules apply.
        return False
    else:
        # We must stringify the moderation action so that it can be
        # stored in the pending request table.
        with _.defer_translation():
            # This will be translated at the point of use.
            reason = _('The message is not from a list member')
        _record_action(msgdata, action.name, sender, reason)
        return True


def _record_action(msgdata, action, sender, reason):
    msgdata['member_moderation_action'] = action
    msgdata['moderation_sender'] = sender
    msgdata.setdefault('moderation_reasons', []).append(reason)


@public
@implementer(IRule)
class NonmemberModeration:
    """The nonmember moderation rule."""

    name = 'nonmember-moderation'
    description = _('Match messages sent by nonmembers.')
    record = True

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        ban_manager = IBanManager(mlist)
        user_manager = getUtility(IUserManager)
        # The NonmemberModeration rule misses unconditionally if any of the
        # senders are banned.
        for sender in msg.senders:
            if ban_manager.is_banned(sender):
                return False
        if len(msg.senders) == 0:
            with _.defer_translation():
                # This will be translated at the point of use.
                reason = _('No sender was found in the message.')
            _record_action(
                msgdata, mlist.default_nonmember_action, 'No sender', reason)
            return True
        # Every sender email must be a member or nonmember directly.  If it is
        # neither, make the email a nonmembers.
        for sender in msg.senders:
            if (mlist.members.get_member(sender) is None
                    and mlist.nonmembers.get_member(sender) is None):   # noqa
                # The email must already be registered, since this happens in
                # the incoming runner itself.
                address = user_manager.get_address(sender)
                assert address is not None, (
                    'Posting address is not registered: {}'.format(sender))
                with suppress(InvalidEmailAddressError):
                    # This might be a list posting address in Reply-To: or
                    # some other invalid address.  In any case, ignore it.
                    mlist.subscribe(address, MemberRole.nonmember)
        # If this message is gated from usenet the rule misses.
        if msgdata.get('fromusenet', False):
            return False
        # Check to see if any of the sender emails is already a member.  If
        # so, then this rule misses.
        member = _find_sender_member(mlist, msg)
        if member is not None:
            return False
        # Do nonmember moderation check.
        for sender in msg.senders:
            nonmember = mlist.nonmembers.get_member(sender)
            assert nonmember is not None, (
                "sender {} didn't get subscribed as a nonmember".format(sender)
                )
            action = nonmember.moderation_action
            if action is not None:
                # Do this first so a specific nonmember.moderation_action
                # trumps the legacy settings.
                return _do_action(msgdata, action, sender)
            # nonmember has no moderation action so check the
            # '*_these_nonmembers' properties.  XXX These are
            # legacy attributes from MM2.1; their database type is 'pickle' and
            # they should eventually get replaced.
            for action_name in ('accept', 'hold', 'reject', 'discard'):
                legacy_attribute_name = '{}_these_nonmembers'.format(
                    action_name)
                checklist = getattr(mlist, legacy_attribute_name)
                for addr in checklist:
                    try:
                        if ((addr.startswith('^') and re.match(addr, sender))
                                or addr == sender):     # noqa: W503
                            # accept_these_nonmembers should 'defer'.
                            if action_name == 'accept':
                                return False
                            with _.defer_translation():
                                # This will be translated at the point of use.
                                reason = (_(
                                    'The sender is in the nonmember {} list'
                                    ), action_name)
                            _record_action(msgdata,
                                           action_name, sender, reason)
                            return True
                    except re.error as error:
                        # The pattern is a malformed regular expression.
                        # Log and continue with the next pattern.
                        log = logging.getLogger('mailman.error')
                        log.error("Invalid regexp '{}' in "
                                  '{}_these_nonmembers for {}: {}'
                                  .format(addr, action_name, mlist.list_id,
                                          error.msg))
                        continue
            # No nonmember.moderation.action and no legacy hits for this
            # sender - continue
        action = mlist.default_nonmember_action
        return _do_action(msgdata, action, sender)



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the retry runner."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.runners.retry import RetryRunner
from mailman.testing.helpers import (
    get_queue_messages,
    make_testable_runner,
    specialized_message_from_string as message_from_string,
)
from mailman.testing.layers import ConfigLayer


class TestRetryRunner(unittest.TestCase):
    """Test the retry runner."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._retryq = config.switchboards['retry']
        self._outq = config.switchboards['out']
        self._runner = make_testable_runner(RetryRunner, 'retry')
        self._msg = message_from_string("""\
From: anne@example.com
To: test@example.com
Message-Id: <first>

""")
        self._msgdata = dict(listid='test.example.com')

    def test_message_put_in_outgoing_queue(self):
        self._retryq.enqueue(self._msg, self._msgdata)
        self._runner.run()
        get_queue_messages('out', expected_count=1)



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Unique ID generation.

Use these functions to create unique ids rather than inlining calls to hashlib
and whatnot.  These are better instrumented for testing purposes.
"""

import os
import time
import uuid
import random
import hashlib

from contextlib import suppress
from flufl.lock import Lock
from mailman.config import config
from mailman.model.uid import UID
from mailman.testing import layers
from public import public


class _PredictableIDGenerator:
    """Base class factory.

    This factory provides a base class for unique ids that need to have
    predictable values in testing mode.
    """

    def __init__(self, context=None):
        # We can't call reset() when the factory is created below, because
        # config.VAR_DIR will not be set at that time.  So initialize it at
        # the first use.
        self._uid_file = None
        self._lock_file = None
        self._lockobj = None
        self._context = context
        layers.MockAndMonkeyLayer.register_reset(self.reset)

    @property
    def _lock(self):
        if self._lockobj is None:
            # These will get automatically cleaned up by the test
            # infrastructure.
            self._uid_file = os.path.join(config.VAR_DIR, '.uid')
            if self._context is not None:
                self._uid_file += '.' + self._context
            self._lock_file = self._uid_file + '.lock'
            self._lockobj = Lock(self._lock_file)
        return self._lockobj

    def new(self):
        """Return a new unique ID or a predictable one if in testing mode."""
        if layers.is_testing():
            # When in testing mode we want to produce predictable ids, but we
            # need to coordinate this among separate processes.  We could use
            # the database, but I don't want to add schema just to handle this
            # case, and besides transactions could get aborted, causing some
            # ids to be recycled.  So we'll use a data file with a lock.  This
            # may still not be ideal due to race conditions, but I think the
            # tests will be serialized enough (and the ids reset between
            # tests) that it will not be a problem.  Maybe.
            return self._next_predictable_id()
        return self._next_unpredictable_id()

    def _next_unpredictable_id(self):
        """Generate a unique id when Mailman is not running in testing mode.

        The type of the returned id is intended to be the type that
        makes sense for the subclass overriding this method.
        """
        raise NotImplementedError

    def _next_predictable_id(self):
        """Generate a predictable id for when Mailman being tested.

        The type of the returned id is intended to be the type that
        makes sense for the subclass overriding this method.
        """
        raise NotImplementedError

    def _next_id(self):
        with self._lock:
            try:
                with open(self._uid_file) as fp:
                    uid = int(fp.read().strip())
                    next_uid = uid + 1              # pragma: no branch
                with open(self._uid_file, 'w') as fp:
                    fp.write(str(next_uid))         # pragma: no branch
                return uid
            except FileNotFoundError:
                with open(self._uid_file, 'w') as fp:
                    fp.write('2')
                return 1

    def reset(self):
        with self._lock:
            with open(self._uid_file, 'w') as fp:
                fp.write('1')


@public
class UIDFactory(_PredictableIDGenerator):
    """A factory for unique ids."""

    def _next_unpredictable_id(self):
        """Return a new UID.

        :return: The new uid
        :rtype: uuid.UUID
        """
        while True:
            uid = uuid.uuid4()
            with suppress(ValueError):
                UID.record(uid)
                return uid

    def _next_predictable_id(self):
        uid = super()._next_id()
        return uuid.UUID(int=uid)


@public
class TokenFactory(_PredictableIDGenerator):

    def __init__(self):
        super().__init__(context='token')

    def _next_unpredictable_id(self):
        """Calculate a unique token.

        Algorithm vetted by the Timbot.  time() has high resolution on
        Linux, clock() on Windows.  random gives us about 45 bits in
        Python 2.2, 53 bits on Python 2.3.  The time and clock values
        basically help obscure the random number generator, as does the
        hash calculation.  The integral parts of the time values are
        discarded because they're the most predictable bits.
        """
        right_now = time.time()
        x = random.random() + right_now % 1.0 + time.process_time() % 1.0
        # Use sha1 because it produces shorter strings.
        return hashlib.sha1(repr(x).encode('utf-8')).hexdigest()

    def _next_predictable_id(self):
        uid = super()._next_id()
        return str(uid).zfill(40)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the protocol support.

For convenience, we currently don't test the http: and https: schemes here.
These are tested fairly well in the template cache tests.  We probably
eventually want to refactor that for test isolation.
"""


import os
import unittest

from contextlib import ExitStack
from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.testing.layers import ConfigLayer
from mailman.utilities import protocols
from tempfile import TemporaryDirectory
from urllib.error import URLError


class TestProtocols(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        resources = ExitStack()
        self.addCleanup(resources.close)
        self.var_dir = resources.enter_context(TemporaryDirectory())
        config.push('template config', """\
        [paths.testing]
        var_dir: {}
        """.format(self.var_dir))
        resources.callback(config.pop, 'template config')
        # Put a demo template in the site directory.
        path = os.path.join(self.var_dir, 'templates', 'site', 'en')
        os.makedirs(path)
        with open(os.path.join(path, 'demo.txt'), 'w') as fp:
            print('Test content', end='', file=fp)
        self._mlist = create_list('test@example.com')

    def test_file(self):
        with TemporaryDirectory() as tempdir:
            path = os.path.join(tempdir, 'my-file')
            with open(path, 'w', encoding='utf-8') as fp:
                print('Some contents', end='', file=fp)
            contents = protocols.get('file:///{}'.format(path))
            self.assertEqual(contents, 'Some contents')

    def test_file_binary(self):
        with TemporaryDirectory() as tempdir:
            path = os.path.join(tempdir, 'my-file')
            with open(path, 'wb') as fp:
                fp.write(b'xxx')
            contents = protocols.get('file:///{}'.format(path), mode='rb')
            self.assertEqual(contents, b'xxx')

    def test_file_ascii(self):
        with TemporaryDirectory() as tempdir:
            path = os.path.join(tempdir, 'my-file')
            with open(path, 'w', encoding='us-ascii') as fp:
                print('Some contents', end='', file=fp)
            contents = protocols.get('file:///{}'.format(path),
                                     encoding='us-ascii')
            self.assertEqual(contents, 'Some contents')

    def test_mailman_internal_uris(self):
        # mailman://demo.txt
        content = protocols.get('mailman:///demo.txt')
        self.assertEqual(content, 'Test content')

    def test_mailman_internal_uris_twice(self):
        # mailman:///demo.txt
        content = protocols.get('mailman:///demo.txt')
        self.assertEqual(content, 'Test content')
        content = protocols.get('mailman:///demo.txt')
        self.assertEqual(content, 'Test content')

    def test_mailman_uri_with_language(self):
        content = protocols.get('mailman:///en/demo.txt')
        self.assertEqual(content, 'Test content')

    def test_mailman_uri_with_english_fallback(self):
        content = protocols.get('mailman:///it/demo.txt')
        self.assertEqual(content, 'Test content')

    def test_mailman_uri_with_list_name(self):
        content = protocols.get('mailman:///test@example.com/demo.txt')
        self.assertEqual(content, 'Test content')

    def test_mailman_full_uri(self):
        content = protocols.get('mailman:///test@example.com/en/demo.txt')
        self.assertEqual(content, 'Test content')

    def test_mailman_full_uri_with_english_fallback(self):
        content = protocols.get('mailman:///test@example.com/it/demo.txt')
        self.assertEqual(content, 'Test content')

    def test_uri_not_found(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///missing.txt')
        self.assertEqual(cm.exception.reason, 'No such file')

    def test_shorter_url_error(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///')
        self.assertEqual(cm.exception.reason, 'No template specified')

    def test_short_url_error(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman://')
        self.assertEqual(cm.exception.reason, 'No template specified')

    def test_bad_language(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///xx/demo.txt')
        self.assertEqual(cm.exception.reason, 'Bad language or list name')

    def test_bad_mailing_list(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///missing@example.com/demo.txt')
        self.assertEqual(cm.exception.reason, 'Bad language or list name')

    def test_missing_mailing_list(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///missing@example.com/it/demo.txt')
        self.assertEqual(cm.exception.reason, 'Missing list')

    def test_no_such_language(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///test@example.com/xx/demo.txt')
        self.assertEqual(cm.exception.reason, 'No such language')

    def test_too_many_path_components(self):
        with self.assertRaises(URLError) as cm:
            protocols.get('mailman:///missing@example.com/en/foo/demo.txt')
        self.assertEqual(cm.exception.reason, 'No such file')

    def test_non_ascii(self):
        # mailman://demo.txt with non-ascii content.
        test_text = b'\xe4\xb8\xad'
        path = os.path.join(self.var_dir, 'templates', 'site', 'it')
        os.makedirs(path)
        with open(os.path.join(path, 'demo.txt'), 'wb') as fp:
            fp.write(test_text)
        content = protocols.get('mailman:///it/demo.txt')
        self.assertIsInstance(content, str)
        self.assertEqual(content, test_text.decode('utf-8'))

    def test_bad_file_keyword(self):
        self.assertRaises(ValueError, protocols.get, 'file:///etc/passwd',
                          invalid_keyword='yes')

    def test_bad_mailman_keyword(self):
        self.assertRaises(ValueError, protocols.get, 'mailman:///demo.text',
                          invalid_keyword='yes')

    def test_bad_protocol(self):
        self.assertRaises(URLError, protocols.get, 'unknown:///demo.text')



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface describing a pipeline handler."""

from public import public
from zope.interface import Attribute, Interface


@public
class IHandler(Interface):
    """A basic pipeline handler."""

    name = Attribute('Handler name; must be unique.')

    description = Attribute('A brief description of the handler.')

    def process(mlist, msg, msgdata):
        """Run the handler.

        :param mlist: The mailing list object.
        :param msg: The message object.
        :param msgdata: The message metadata.
        """



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Building blocks for styles.

Use these to compose higher level styles.  Their apply() methods deliberately
have no super() upcall.  You need to explicitly call all base class apply()
methods in your compositional derived class.
"""


from datetime import timedelta
from mailman.core.i18n import _
from mailman.interfaces.action import Action, FilterAction
from mailman.interfaces.archiver import ArchivePolicy
from mailman.interfaces.autorespond import ResponseAction
from mailman.interfaces.bounce import UnrecognizedBounceDisposition
from mailman.interfaces.digests import DigestFrequency
from mailman.interfaces.mailinglist import (
    ArchiveRenderingMode,
    DMARCMitigateAction,
    Personalization,
    ReplyToMunging,
    SubscriptionPolicy,
)
from mailman.interfaces.nntp import NewsgroupModeration
from mailman.model.roster import RosterVisibility
from public import public


@public
class Identity:
    """Set basic identify attributes."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        mlist.display_name = mlist.list_name.capitalize()
        mlist.include_rfc2369_headers = True
        mlist.volume = 1
        mlist.post_id = 1
        mlist.description = ''
        mlist.info = ''
        mlist.preferred_language = 'en'
        mlist.subject_prefix = _('[${mlist.display_name}] ')
        mlist.encode_ascii_prefixes = (
            mlist.preferred_language.charset != 'us-ascii')


@public
class BasicOperation:
    """Set basic operational attributes."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        mlist.emergency = False
        mlist.personalize = Personalization.none
        mlist.default_member_action = Action.defer
        mlist.default_nonmember_action = Action.hold
        mlist.subscription_policy = SubscriptionPolicy.confirm
        mlist.unsubscription_policy = SubscriptionPolicy.confirm
        mlist.member_roster_visibility = RosterVisibility.moderators
        # Notify the administrator of pending requests and membership changes.
        mlist.admin_immed_notify = True
        mlist.admin_notify_mchanges = False
        mlist.respond_to_post_requests = True
        mlist.obscure_addresses = True
        mlist.collapse_alternatives = True
        mlist.convert_html_to_plaintext = False
        mlist.filter_action = FilterAction.discard
        mlist.filter_content = False
        # Digests.
        mlist.digests_enabled = True
        mlist.digest_is_default = False
        mlist.digest_size_threshold = 30          # KB
        mlist.digest_send_periodic = True
        mlist.digest_volume_frequency = DigestFrequency.monthly
        mlist.next_digest_number = 1
        # DMARC
        mlist.dmarc_mitigate_action = DMARCMitigateAction.no_mitigation
        mlist.dmarc_mitigate_unconditionally = False
        mlist.dmarc_moderation_notice = ''
        mlist.dmarc_wrapped_message_text = ''
        # NNTP gateway
        mlist.linked_newsgroup = ''
        mlist.gateway_to_news = False
        mlist.gateway_to_mail = False
        mlist.nntp_prefix_subject_too = True
        # In patch #401270, this was called newsgroup_is_moderated, but the
        # semantics weren't quite the same.
        mlist.newsgroup_moderation = NewsgroupModeration.none
        # Topics
        #
        # `topics' is a list of 4-tuples of the following form:
        #
        #     (name, pattern, description, emptyflag)
        #
        # name is a required arbitrary string displayed to the user when they
        # get to select their topics of interest
        #
        # pattern is a required verbose regular expression pattern which is
        # used as IGNORECASE.
        #
        # description is an optional description of what this topic is
        # supposed to match
        #
        # emptyflag is a boolean used internally in the admin interface to
        # signal whether a topic entry is new or not (new ones which do not
        # have a name or pattern are not saved when the submit button is
        # pressed).
        mlist.topics = []
        mlist.topics_enabled = False
        mlist.topics_bodylines_limit = 5
        # This is a mapping between user "names" (i.e. addresses) and
        # information about which topics that user is interested in.  The
        # values are a list of topic names that the user is interested in,
        # which should match the topic names in mlist.topics above.
        #
        # If the user has not selected any topics of interest, then the rule
        # is that they will get all messages, and they will not have an entry
        # in this dictionary.
        mlist.topics_userinterest = {}
        # scrub regular delivery
        mlist.scrub_nondigest = False
        # Archive rendering mode.
        mlist.archive_rendering_mode = ArchiveRenderingMode.text


@public
class Bounces:
    """Basic bounce processing."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        # Bounces
        mlist.forward_unrecognized_bounces_to = (
            UnrecognizedBounceDisposition.administrators)
        mlist.process_bounces = True
        mlist.bounce_score_threshold = 5.0
        mlist.bounce_info_stale_after = timedelta(days=7)
        mlist.bounce_you_are_disabled_warnings = 3
        mlist.bounce_you_are_disabled_warnings_interval = timedelta(days=7)
        mlist.bounce_notify_owner_on_bounce_increment = False
        mlist.bounce_notify_owner_on_disable = True
        mlist.bounce_notify_owner_on_removal = True
        # Autoresponder
        mlist.autorespond_owner = ResponseAction.none
        mlist.autoresponse_owner_text = ''
        mlist.autorespond_postings = ResponseAction.none
        mlist.autoresponse_postings_text = ''
        mlist.autorespond_requests = ResponseAction.none
        mlist.autoresponse_request_text = ''
        mlist.autoresponse_grace_period = timedelta(days=90)
        # This holds legacy member related information.  It's keyed by the
        # member address, and the value is an object containing the bounce
        # score, the date of the last received bounce, and a count of the
        # notifications left to send.
        mlist.bounce_info = {}
        # New style delivery status
        mlist.delivery_status = {}
        # The processing chain that messages posted to this mailing list get
        # processed by.
        mlist.posting_chain = 'default-posting-chain'
        # The default pipeline to send accepted messages through to the
        # mailing list's members.
        mlist.posting_pipeline = 'default-posting-pipeline'
        # The processing chain that messages posted to this mailing list's
        # -owner address gets processed by.
        mlist.owner_chain = 'default-owner-chain'
        # The default pipeline to send -owner email through.
        mlist.owner_pipeline = 'default-owner-pipeline'


@public
class Public:
    """Settings for public mailing lists."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        mlist.advertised = True
        mlist.reply_goes_to_list = ReplyToMunging.no_munging
        mlist.reply_to_address = ''
        mlist.first_strip_reply_to = False
        mlist.archive_policy = ArchivePolicy.public


@public
class Private:
    """Settings for private mailing lists."""

    def apply(self, mailing_list):
        mlist = mailing_list
        mlist.advertised = False
        mlist.archive_policy = ArchivePolicy.private
        mlist.subscription_policy = SubscriptionPolicy.confirm_then_moderate


@public
class Announcement:
    """Settings for announce-only lists."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        mlist.allow_list_posts = False
        mlist.send_welcome_message = True
        mlist.send_goodbye_message = True
        mlist.anonymous_list = False


@public
class Discussion:
    """Settings for standard discussion lists."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        mlist.allow_list_posts = True
        mlist.send_welcome_message = True
        mlist.send_goodbye_message = True
        mlist.anonymous_list = False


@public
class Moderation:
    """Settings for basic moderation."""

    def apply(self, mailing_list):
        # For cut-n-paste convenience.
        mlist = mailing_list
        mlist.max_num_recipients = 10
        mlist.max_message_size = 40               # KB
        mlist.require_explicit_destination = True
        mlist.bounce_matching_headers = """
# Lines that *start* with a '#' are comments.
to: friend@public.com
message-id: relay.comanche.denmark.eu
from: list@listme.com
from: .*@uplinkpro.com
"""
        mlist.header_matches = []
        mlist.administrivia = True
        # Member moderation.
        mlist.member_moderation_notice = ''
        mlist.accept_these_nonmembers = []
        mlist.hold_these_nonmembers = []
        mlist.reject_these_nonmembers = []
        mlist.discard_these_nonmembers = []
        mlist.forward_auto_discards = True
        mlist.nonmember_rejection_notice = ''
        # automatic discarding
        mlist.max_days_to_hold = 0



# Copyright (C) 2002-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The email commands 'end' and 'stop'."""

from mailman.core.i18n import _
from mailman.interfaces.command import ContinueProcessing, IEmailCommand
from public import public
from zope.interface import implementer


@public
@implementer(IEmailCommand)
class End:
    """The email 'end' command."""

    name = 'end'
    argument_description = ''
    description = _('Stop processing commands.')
    short_description = description

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        # Ignore all arguments.
        return ContinueProcessing.no


@public
class Stop(End):
    """The email 'stop' command (an alias for 'end')."""

    name = 'stop'
    description = _("An alias for 'end'.")
    short_description = description



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test option utilities."""

import click
import unittest

from mailman.utilities.options import validate_runner_spec


class TestValidateRunnerSpec(unittest.TestCase):
    def test_false_value(self):
        self.assertIsNone(validate_runner_spec(None, None, None))

    def test_runner_only(self):
        specs = validate_runner_spec(None, None, 'incoming')
        self.assertEqual(specs, ('incoming', 1, 1))

    def test_full_runner_spec(self):
        specs = validate_runner_spec(None, None, 'incoming:2:4')
        self.assertEqual(specs, ('incoming', 2, 4))

    def test_bad_runner_spec(self):
        with self.assertRaises(click.BadParameter) as cm:
            validate_runner_spec(None, None, 'incoming:not:int')
        self.assertEqual(
            cm.exception.message,
            'slice and range must be integers: incoming:not:int')

    def test_bad_runner_spec_parts(self):
        with self.assertRaises(click.UsageError) as cm:
            validate_runner_spec(None, None, 'incoming:2')
        self.assertEqual(
            cm.exception.message,
            'Bad runner spec: incoming:2')



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `mailman members` command."""

import unittest

from click.testing import CliRunner
from mailman.app.lifecycle import create_list
from mailman.commands.cli_delmembers import delmembers
from mailman.testing.helpers import get_queue_messages, subscribe
from mailman.testing.layers import ConfigLayer
from tempfile import NamedTemporaryFile


class TestCLIDelMembers(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        # Default to no messages.
        self._mlist.send_welcome_message = False
        self._mlist.send_goodbye_message = False
        self._mlist.admin_notify_mchanges = False
        self._command = CliRunner()

    def test_no_such_list(self):
        result = self._command.invoke(delmembers, (
            '-f', '-', '-l', 'bee.example.com'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: No such list: bee.example.com\n')

    def test_bad_filename(self):
        result = self._command.invoke(delmembers, (
            '-f', 'bad', '-l', 'ant.example.com'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: Invalid value for \'--file\' / \'-f\': '
            '\'bad\': No such file or directory\n')

    def test_not_subscribed_without_display_name(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(
           result.output,
           'Member not subscribed (skipping): aperson@example.com\n'
           )

    def test_not_subscribed_with_display_name(self):
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            result = self._command.invoke(delmembers, (
                '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(
           result.output,
           'Member not subscribed (skipping): '
           'Anne Person <aperson@example.com>\n'
           )

    def test_deletion_blank_lines(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        subscribe(self._mlist, 'Cate')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('', file=infp)
            print('   ', file=infp)
            print('\t', file=infp)
            print('Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(delmembers, (
                '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 1)
        self.assertEqual(str(members[0].address),
                         'Cate Person <cperson@example.com>')

    def test_deletion_commented_lines(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Anne Person <aperson@example.com>', file=infp)
            print('#Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(delmembers, (
                '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 1)
        self.assertEqual(str(members[0].address),
                         'Bart Person <bperson@example.com>')

    def test_deletion_input_stdin(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        result = self._command.invoke(delmembers, (
            '-f', '-', '-l', 'ant.example.com'),
            input='Anne Person <aperson@example.com>\n')
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 1)
        self.assertEqual(str(members[0].address),
                         'Bart Person <bperson@example.com>')

    def test_deletion_all_members(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        result = self._command.invoke(delmembers, (
            '-a', '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)

    def test_deletion_from_all_lists(self):
        self._mlist2 = create_list('bee@example.com')
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        subscribe(self._mlist2, 'Bart')
        subscribe(self._mlist2, 'Cate')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(delmembers, (
                '--fromall', '-f', infp.name))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 1)
        self.assertEqual(str(members[0].address),
                         'Anne Person <aperson@example.com>')
        members2 = list(self._mlist2.members.members)
        self.assertEqual(len(members2), 1)
        self.assertEqual(str(members2[0].address),
                         'Cate Person <cperson@example.com>')

    def test_deletion_from_all_lists_not_member_of_all(self):
        self._mlist2 = create_list('bee@example.com')
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist2, 'Bart')
        subscribe(self._mlist2, 'Cate')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(delmembers, (
                '--fromall', '-f', infp.name))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 1)
        self.assertEqual(str(members[0].address),
                         'Anne Person <aperson@example.com>')
        members2 = list(self._mlist2.members.members)
        self.assertEqual(len(members2), 1)
        self.assertEqual(str(members2[0].address),
                         'Cate Person <cperson@example.com>')

    def test_deletion_dash_member(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        result = self._command.invoke(delmembers, (
            '-m', 'aperson@example.com', '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 1)
        self.assertEqual(str(members[0].address),
                         'Bart Person <bperson@example.com>')

    def test_deletion_multiple_dash_member(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        result = self._command.invoke(delmembers, (
            '-m', 'aperson@example.com', '-l', 'ant.example.com',
            '-m', 'bperson@example.com'))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)

    def test_deletion_file_and_dash_member(self):
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlist, 'Bart')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('Bart Person <bperson@example.com>', file=infp)
            result = self._command.invoke(delmembers, (
                '-m', 'aperson@example.com', '-l', 'ant.example.com',
                '-f', infp.name))
        self.assertEqual(result.output, '')
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)

    def test_invalid_fromall_list(self):
        result = self._command.invoke(delmembers, (
            '-f', '-', '--fromall', '-l', 'bee.example.com'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: --fromall may not be specified with -l/--list, '
            'or -a/--all\n')

    def test_invalid_fromall_all(self):
        result = self._command.invoke(delmembers, (
            '--fromall', '-a'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: --fromall may not be specified with -l/--list, '
            'or -a/--all\n')

    def test_invalid_file_all(self):
        result = self._command.invoke(delmembers, (
            '-f', '-', '-a'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: -a/--all must not be specified with '
            '-f/--file or -m/--member.\n')

    def test_invalid_member_all(self):
        result = self._command.invoke(delmembers, (
            '-m', 'aperson@example.com', '-a'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: -a/--all must not be specified with '
            '-f/--file or -m/--member.\n')

    def test_invalid_not_fromall_not_list(self):
        result = self._command.invoke(delmembers, (
            '-f', '-'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: Without --fromall, -l/--list is required.\n')

    def test_invalid_no_all_file_or_member(self):
        result = self._command.invoke(delmembers, (
            '-l', 'ant@example.com'))
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(
            result.output,
            'Usage: delmembers [OPTIONS]\n'
            'Try \'delmembers --help\' for help.\n\n'
            'Error: At least one of -a/--all, -f/--file or -m/--member '
            'is required.\n')

    def test_override_no_goodbye(self):
        self._mlist.send_goodbye_message = False
        subscribe(self._mlist, 'Anne')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-g', '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)
        items = get_queue_messages('virgin', expected_count=1)
        self.assertIn('You have been unsubscribed',
                      str(items[0].msg['subject']))

    def test_override_yes_goodbye(self):
        self._mlist.send_goodbye_message = True
        subscribe(self._mlist, 'Anne')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-G', '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)
        get_queue_messages('virgin', expected_count=0)

    def test_no_override_goodbye(self):
        self._mlist.send_goodbye_message = True
        subscribe(self._mlist, 'Anne')
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)
        items = get_queue_messages('virgin', expected_count=1)
        self.assertIn('You have been unsubscribed',
                      str(items[0].msg['subject']))

    def test_override_no_admin_notify(self):
        subscribe(self._mlist, 'Anne')
        self._mlist.admin_notify_mchanges = False
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-n', '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)
        items = get_queue_messages('virgin', expected_count=1)
        self.assertIn('Ant unsubscription notification',
                      str(items[0].msg['subject']))
        self.assertIn('Anne Person <aperson@example.com> has been removed',
                      str(items[0].msg))

    def test_override_yes_admin_notify(self):
        subscribe(self._mlist, 'Anne')
        self._mlist.admin_notify_mchanges = True
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-N', '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)
        get_queue_messages('virgin', expected_count=0)

    def test_no_override_admin_notify(self):
        subscribe(self._mlist, 'Anne')
        self._mlist.admin_notify_mchanges = True
        with NamedTemporaryFile('w', buffering=1, encoding='utf-8') as infp:
            print('aperson@example.com', file=infp)
            result = self._command.invoke(delmembers, (
                '-f', infp.name, '-l', 'ant.example.com'))
        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)
        members = list(self._mlist.members.members)
        self.assertEqual(len(members), 0)
        items = get_queue_messages('virgin', expected_count=1)
        self.assertIn('Ant unsubscription notification',
                      str(items[0].msg['subject']))
        self.assertIn('Anne Person <aperson@example.com> has been removed',
                      str(items[0].msg))



"""Members and list moderation action.

Revision ID: 7b254d88f122
Revises: d4fbb4fd34ca
Create Date: 2016-02-10 11:31:04.233619

This is a data-only migration.  If a member has the same moderation action as
the mailing list's default, then set its moderation action to None and use the
fallback to the list's default.
"""


import sqlalchemy as sa

from alembic import op
from mailman.database.types import Enum, SAUnicode
from mailman.interfaces.action import Action
from mailman.interfaces.member import MemberRole


# Revision identifiers, used by Alembic.
revision = '7b254d88f122'
down_revision = 'd4fbb4fd34ca'


mailinglist_table = sa.sql.table(
    'mailinglist',
    sa.sql.column('id', sa.Integer),
    sa.sql.column('list_id', SAUnicode),
    sa.sql.column('default_member_action', Enum(Action)),
    sa.sql.column('default_nonmember_action', Enum(Action)),
    )


member_table = sa.sql.table(
    'member',
    sa.sql.column('id', sa.Integer),
    sa.sql.column('list_id', SAUnicode),
    sa.sql.column('role', Enum(MemberRole)),
    sa.sql.column('moderation_action', Enum(Action)),
    )


# This migration only considers members and nonmembers.
members_query = member_table.select().where(sa.or_(
    member_table.c.role == MemberRole.member,
    member_table.c.role == MemberRole.nonmember,
    ))


# list-id -> {property-name -> action}
#
# where property-name will be either default_member_action or
# default_nonmember_action.
DEFAULT_ACTION_CACHE = {}
MISSING = object()


def _get_default_action(connection, member):
    # breakpoint()
    list_id = member[1]
    property_name = 'default_{}_action'.format(member[2].name)
    list_mapping = DEFAULT_ACTION_CACHE.setdefault(list_id, {})
    action = list_mapping.get(property_name, MISSING)
    if action is MISSING:
        action = (
            connection.execute(
                sa.select(getattr(mailinglist_table.c, property_name))
                .where(mailinglist_table.c.list_id == list_id))
            .fetchone()[0]
            )
        list_mapping[property_name] = action
    return action


def upgrade():
    connection = op.get_bind()
    for member in connection.execute(members_query).fetchall():
        default_action = _get_default_action(connection, member)
        # If the (non)member's moderation action is the same as the mailing
        # list's default, then set it to None.  The moderation rule will
        # fallback to the list's default.
        if member[3] == default_action:
            connection.execute(member_table.update().where(
                member_table.c.id == member[0]
                ).values(moderation_action=None))


def downgrade():
    connection = op.get_bind()
    for member in connection.execute(members_query.where(
           member_table.c.moderation_action == None)).fetchall():  # noqa: E711
        default_action = _get_default_action(connection, member)
        # Use the mailing list's default action
        connection.execute(member_table.update().where(
            member_table.c.id == member[0]
            ).values(moderation_action=default_action))



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The 'confirm' email command."""

from mailman.app.moderator import handle_message
from mailman.core.i18n import _
from mailman.interfaces.action import Action
from mailman.interfaces.command import ContinueProcessing, IEmailCommand
from mailman.interfaces.member import (
    AlreadySubscribedError,
    MembershipIsBannedError,
)
from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import ISubscriptionManager, TokenOwner
from mailman.rules.approved import Approved
from public import public
from zope.component import getUtility
from zope.interface import implementer


@public
@implementer(IEmailCommand)
class Confirm:
    """The email 'confirm' command."""

    name = 'confirm'
    argument_description = 'token'
    description = _('Confirm a subscription or held message request.')
    short_description = description

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        # The token must be in the arguments.
        if len(arguments) == 0:
            print(_('No confirmation token found'), file=results)
            return ContinueProcessing.no
        # Make sure we don't try to confirm the same token more than once.
        token = arguments[0]
        tokens = getattr(results, 'confirms', set())
        if token in tokens:
            # Do not try to confirm this one again.
            return ContinueProcessing.no
        tokens.add(token)
        results.confirms = tokens
        # now figure out what this token is for.
        pendable = getUtility(IPendings).confirm(token, expunge=False)
        if pendable is None:
            print(_('Confirmation token did not match'), file=results)
            results.send_response = True
            return ContinueProcessing.no
        if pendable['type'] == 'held message':
            return self.process_held(
                mlist, msg, msgdata, token, pendable, results)
        try:
            new_token, token_owner, member = ISubscriptionManager(
                mlist).confirm(token)
            if new_token is None:
                assert token_owner is TokenOwner.no_one, token_owner
                # We can't assert anything about member.  It will be None when
                # the workflow we're confirming is a subscription request,
                # and non-None when we're confirming an unsubscription request.
                # This class doesn't know which is happening.
                succeeded = True
            elif token_owner is TokenOwner.moderator:
                # This must have been a confirm-then-moderate (un)subscription.
                assert new_token != token
                # We can't assert anything about member for the above reason.
                succeeded = True
            else:
                assert token_owner is not TokenOwner.no_one, token_owner
                assert member is None, member
                succeeded = False
        except LookupError:
            # The token must not exist in the database.
            succeeded = False
        except (MembershipIsBannedError, AlreadySubscribedError) as e:
            print(str(e), file=results)
            return ContinueProcessing.no
        if succeeded:
            print(_('Confirmed'), file=results)
            # After the 'confirm' command, do not process any other commands in
            # the email.
            return ContinueProcessing.no
        print(_('Confirmation token did not match'), file=results)
        return ContinueProcessing.no

    def process_held(self, mlist, msg, msgdata, token, pendable, results):
        # Is this confirmation approved?
        approved = Approved().check(mlist, msg, msgdata)
        if approved:
            action = Action.accept
            message = 'accepted'                          # noqa: F841
        elif not msgdata.get('has_approved', False):
            action = Action.discard
            message = 'discarded'                         # noqa: F841
        else:
            print(_('Invalid Approved: password'), file=results)
            return ContinueProcessing.no
        handle_message(mlist, pendable['id'], action)
        print(_('Message ${message}'), file=results)
        return ContinueProcessing.no



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Perform origination & content authentication checks and add
an Authentication-Results header to the outgoing message"""

import logging

from authheaders import authenticate_message
from authres import AuthenticationResultsHeader, SyntaxError
from dns.resolver import Timeout
from itertools import chain
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.handler import IHandler
from mailman.utilities.retry import retry
from public import public
from zope.interface import implementer


log = logging.getLogger('mailman.debug')

# Number of times to retry authentication checks in case of DNS failure.
NUM_TIMEOUT_RETRIES = 2

# This value is used by the test suite to provide a faux DNS resolver.
dnsfunc = None

# Email header including the results of ARC validation from the sender.
AUTH_RESULT_HEADER = 'Authentication-Results'


def prepend_headers(msg, headers):
    """Appends a group of headers to the beginning of a message.

    :param msg: The message object to add headers to.
    :type msg: email.message.EmailMessage
    :param headers: The list of headers to added to message.
    :type msg: List(email.headers.Header)
    """
    old_headers = msg.items()

    for key in msg:
        del msg[key]

    for k, v in chain(headers, old_headers):
        msg[k] = v


def trusted_auth_res(msg):
    """Extract the most recent trusted Authentication-Results(AR) header.

    :param msg: The message to extract the AR header from.
    :type msg: email.message.EmailMessage.
    """

    if config.arc.trusted_authserv_ids and (AUTH_RESULT_HEADER in msg):
        header = '{}: {}'.format(AUTH_RESULT_HEADER, msg[AUTH_RESULT_HEADER])
        # prod.outlook.com sends malformed headers. If we find these,
        # ignore them, as they are the equivalent of not trusted.
        # See https://gitlab.com/mailman/mailman/-/issues/885
        try:
            authserv_id = AuthenticationResultsHeader.parse(header).authserv_id
            if authserv_id in config.arc.trusted_authserv_ids:
                return header
        except SyntaxError:
            # Treat the faulty header as though it does not exist
            return


@retry(Timeout, NUM_TIMEOUT_RETRIES)
def authenticate(msg, msgdata):
    """ARC verify a message and update the Authentication-Results header.

    If there is a previous Authentication-Results, remove that and add a
    new one.
    """
    prev = trusted_auth_res(msg)
    auth_result = authenticate_message(
        msg.as_bytes(), config.arc.authserv_id,
        prev=prev,
        spf=False,  # cant spf check in mailman
        dkim=config.arc.dkim_enabled,
        dmarc=config.arc.dmarc_enabled,
        arc=True,
        dnsfunc=dnsfunc)

    if AUTH_RESULT_HEADER in msg:
        del msg[AUTH_RESULT_HEADER]

    auth_result = auth_result.split(':', 1)[1].strip()
    prepend_headers(msg, [(AUTH_RESULT_HEADER, auth_result)])


@public
@implementer(IHandler)
class ValidateAuthenticity:
    """Perform authentication checks and attach resulting headers.

    Validate the ARC chain of a message and add results to a new
    Authentication-Results header
    """

    name = 'validate-authenticity'
    description = _(
        """Perform auth checks and attach Authentication-Results header.""")

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        if config.arc_enabled:
            authenticate(msg, msgdata)



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test some additional plugin stuff."""

import sys
import unittest

from contextlib import ExitStack
from mailman.core.initialize import run_plugin_pre_hook
from mailman.interfaces.plugin import IPlugin
from mailman.plugins.initialize import initialize
from mailman.plugins.testing.layer import PluginRESTLayer
from mailman.testing.helpers import call_api
from mailman.testing.layers import ConfigLayer
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from unittest.mock import patch
from urllib.error import HTTPError
from zope.interface import implementer


class TestRESTPlugin(unittest.TestCase):
    layer = PluginRESTLayer

    def test_plugin_raises_exception(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/example/no')
        self.assertEqual(cm.exception.code, 400)


@implementer(IPlugin)
class TestablePlugin:
    def pre_hook(self):
        pass

    def post_hook(self):
        pass

    resource = None


@implementer(IPlugin)
class ExceptionRaisingPlugin:
    def pre_hook(self):
        raise Exception('Bad plugin')

    def post_hook(self):
        pass
    resource = None


class TestInitializePlugins(unittest.TestCase):
    layer = ConfigLayer

    def test_duplicate_plugin_name(self):
        with ExitStack() as resources:
            system_path = resources.enter_context(TemporaryDirectory())
            fake_plugin_config = {
                'path': system_path,
                'enabled': 'yes',
                'class': 'ExamplePlugin',
                }
            log_mock = resources.enter_context(
                patch('mailman.plugins.initialize.log'))
            fake_mailman_config = SimpleNamespace(
                plugin_configs=[('example', fake_plugin_config)],
                plugins=['example'],
                )
            resources.enter_context(patch(
                'mailman.plugins.initialize.config', fake_mailman_config))
            initialize()
            log_mock.error.assert_called_once_with(
                'Duplicate plugin name: example')

    def test_does_not_implement(self):
        with ExitStack() as resources:
            system_path = resources.enter_context(TemporaryDirectory())
            fake_plugin_config = {
                'path': system_path,
                'enabled': 'yes',
                'class': 'ExamplePlugin',
                }
            log_mock = resources.enter_context(
                patch('mailman.plugins.initialize.log'))
            fake_mailman_config = SimpleNamespace(
                plugin_configs=[('example', fake_plugin_config)],
                plugins=[],
                )
            resources.enter_context(patch(
                'mailman.plugins.initialize.config', fake_mailman_config))
            resources.enter_context(patch(
                'mailman.plugins.initialize.call_name',
                # object() does not implement IPlugin.
                return_value=object()))
            initialize()
            log_mock.error.assert_called_once_with(
                'Plugin class does not implement IPlugin: ExamplePlugin')
            self.assertNotIn(system_path, sys.path)

    def test_adds_plugins_to_config(self):
        with ExitStack() as resources:
            system_path = resources.enter_context(TemporaryDirectory())
            fake_plugin_config = {
                'path': system_path,
                'enabled': 'yes',
                'class': 'ExamplePlugin',
                }
            fake_mailman_config = SimpleNamespace(
                plugin_configs=[('example', fake_plugin_config)],
                plugins={},
                )
            resources.enter_context(patch(
                'mailman.plugins.initialize.config', fake_mailman_config))
            testable_plugin = TestablePlugin()
            resources.enter_context(patch(
                'mailman.plugins.initialize.call_name',
                # object() does not implement IPlugin.
                return_value=testable_plugin))
            initialize()
            self.assertIn('example', fake_mailman_config.plugins)
            self.assertEqual(
                fake_mailman_config.plugins['example'],
                testable_plugin)

    def test_not_enabled(self):
        with ExitStack() as resources:
            fake_plugin_config = {
                'path': '/does/not/exist',
                'enabled': 'no',
                'class': 'ExamplePlugin',
                }
            log_mock = resources.enter_context(
                patch('mailman.plugins.initialize.log'))
            fake_mailman_config = SimpleNamespace(
                plugin_configs=[('example', fake_plugin_config)],
                plugins={},
                )
            resources.enter_context(patch(
                'mailman.plugins.initialize.config', fake_mailman_config))
            initialize()
            log_mock.info.assert_called_once_with(
                'Plugin not enabled, or empty class path: example')

    def test_plugin_disabled_with_exception(self):
        with ExitStack() as resources:
            system_path = resources.enter_context(TemporaryDirectory())
            fake_plugin_config = {
                'path': system_path,
                'enabled': 'yes',
                'class': 'ExamplePlugin',
                }
            testable_plugin = ExceptionRaisingPlugin()
            fake_mailman_config = SimpleNamespace(
                plugin_configs=[('example', fake_plugin_config)],
                plugins={'example': testable_plugin},
                )

            resources.enter_context(patch(
                'mailman.plugins.initialize.call_name',
                return_value=testable_plugin))
            run_plugin_pre_hook(fake_mailman_config)
            self.assertNotIn('example', fake_mailman_config.plugins)



# Copyright (C) 1998-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Handler for automatic responses."""

import logging

from mailman.core.i18n import _
from mailman.email.message import UserNotification
from mailman.interfaces.autorespond import (
    ALWAYS_REPLY,
    IAutoResponseSet,
    Response,
    ResponseAction,
)
from mailman.interfaces.handler import IHandler
from mailman.interfaces.usermanager import IUserManager
from mailman.utilities.datetime import today
from mailman.utilities.string import expand, wrap
from public import public
from zope.component import getUtility
from zope.interface import implementer


log = logging.getLogger('mailman.error')


@public
@implementer(IHandler)
class Replybot:
    """Send automatic responses."""

    name = 'replybot'
    description = _('Send automatic responses.')

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        # There are several cases where the replybot is short-circuited:
        # * the original message has an "X-Ack: No" header
        # * the message has a Precedence header with values bulk, junk, or
        #   list, and there's no explicit "X-Ack: yes" header
        # * the message metadata has a true 'noack' key
        ack = msg.get('x-ack', '').lower()
        if ack == 'no' or msgdata.get('noack'):
            return
        precedence = msg.get('precedence', '').lower()
        if ack != 'yes' and precedence in ('bulk', 'junk', 'list'):
            return
        # Check to see if the list is even configured to autorespond to this
        # email message.  Note: the incoming message processors should set the
        # destination key in the message data.
        if msgdata.get('to_owner'):
            if mlist.autorespond_owner is ResponseAction.none:
                return
            response_type = Response.owner
            response_text = mlist.autoresponse_owner_text
        elif msgdata.get('to_request'):
            if mlist.autorespond_requests is ResponseAction.none:
                return
            response_type = Response.command
            response_text = mlist.autoresponse_request_text
        elif msgdata.get('to_list'):
            if mlist.autorespond_postings is ResponseAction.none:
                return
            response_type = Response.postings
            response_text = mlist.autoresponse_postings_text
        else:
            # There are no automatic responses for any other destination.
            return
        # Now see if we're in the grace period for this sender.  grace_period
        # = 0 means always automatically respond, as does an "X-Ack: yes"
        # header (useful for debugging).
        response_set = IAutoResponseSet(mlist)
        user_manager = getUtility(IUserManager)
        address = user_manager.get_address(msg.sender)
        if address is None:
            address = user_manager.create_address(msg.sender)
        grace_period = mlist.autoresponse_grace_period
        if grace_period > ALWAYS_REPLY and ack != 'yes':
            last = response_set.last_response(address, response_type)
            if last is not None and last.date_sent + grace_period > today():
                return
        # Okay, we know we're going to respond to this sender, craft the
        # message, send it, and update the database.
        display_name = mlist.display_name
        subject = _(
            'Auto-response for your message to the "${display_name}" '
            'mailing list')
        # Do string interpolation into the autoresponse text
        d = dict(
            list_name=mlist.list_name,
            display_name=display_name,
            requestemail=mlist.request_address,
            owneremail=mlist.owner_address,
            )
        # Interpolation and Wrap the response text.
        text = wrap(expand(response_text, mlist, d))
        outmsg = UserNotification(msg.sender, mlist.bounces_address,
                                  subject, text, mlist.preferred_language)
        outmsg['X-Mailer'] = _('The Mailman Replybot')
        # prevent recursions and mail loops!
        outmsg['X-Ack'] = 'No'
        outmsg.send(mlist)
        response_set.response_sent(address, response_type)



# Copyright (C) 2013-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Print the mailman configuration."""

import click

from lazr.config._config import Section
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.command import ICLISubCommand
from mailman.utilities.options import I18nCommand
from public import public
from zope.interface import implementer


def _section_exists(section):
    # Not all of the attributes in config are actual sections, so we have to
    # check the section's type.
    return (
        hasattr(config, section) and
        isinstance(getattr(config, section), Section)
        )


def _get_value(section, key):
    return getattr(getattr(config, section), key)


def _print_values_for_section(section, output):
    current_section = sorted(getattr(config, section))
    for key in current_section:
        value = _get_value(section, key)
        print('[{}] {}: {}'.format(section, key, value), file=output)


@click.command(
    cls=I18nCommand,
    help=_('Print the Mailman configuration.'))
@click.option(
    '--output', '-o',
    type=click.File(mode='w', encoding='utf-8', atomic=True),
    help=_("""\
    File to send the output to.  If not given, or if '-' is given, standard
    output is used."""))
@click.option(
    '--section', '-s',
    help=_("""\
    Section to use for the lookup.  If no key is given, all the key-value pairs
    of the given section will be displayed."""))
@click.option(
    '--key', '-k',
    help=_("""\
    Key to use for the lookup.  If no section is given, all the key-values pair
    from any section matching the given key will be displayed."""))
@click.pass_context
def conf(ctx, output, section, key):
    # Case 1: Both section and key are given, so we can look the value up
    # directly.
    if section is not None and key is not None:
        if not _section_exists(section):
            ctx.fail('No such section: {}'.format(section))
        elif not hasattr(getattr(config, section), key):
            ctx.fail('Section {}: No such key: {}'.format(section, key))
        else:
            print(_get_value(section, key), file=output)
    # Case 2: Section is given, key is not given.
    elif section is not None and key is None:
        if _section_exists(section):
            _print_values_for_section(section, output)
        else:
            ctx.fail('No such section: {}'.format(section))
    # Case 3: Section is not given, key is given.
    elif section is None and key is not None:
        for current_section in sorted([section.name for section in config]):
            # We have to ensure that the current section actually exists
            # and that it contains the given key.
            if (_section_exists(current_section) and
                    hasattr(getattr(config, current_section), key)):
                value = _get_value(current_section, key)
                print('[{}] {}: {}'.format(
                    current_section, key, value), file=output)
    # Case 4: Neither section nor key are given, just display all the
    # sections and their corresponding key/value pairs.
    elif section is None and key is None:
        for current_section in sorted([section.name for section in config]):
            # However, we have to make sure that the current sections and
            # key which are being looked up actually exist before trying
            # to print them.
            if _section_exists(current_section):
                _print_values_for_section(current_section, output)
    else:
        raise AssertionError('Unexpected combination')


@public
@implementer(ICLISubCommand)
class Conf:
    name = 'conf'
    command = conf



"""List subscription policy

Revision ID: 16c2b25c7b
Revises: 46e92facee7
Create Date: 2015-03-21 11:00:44.634883

"""

import sqlalchemy as sa

from alembic import op
from mailman.database.helpers import exists_in_db, is_sqlite
from mailman.database.types import Enum
from mailman.interfaces.mailinglist import SubscriptionPolicy


# Revision identifiers, used by Alembic.
revision = '16c2b25c7b'
down_revision = '46e92facee7'


def upgrade():
    # Update the schema.
    if not exists_in_db(op.get_bind(), 'mailinglist', 'subscription_policy'):
        # SQLite may not have removed it when downgrading.
        op.add_column('mailinglist', sa.Column(
            'subscription_policy', Enum(SubscriptionPolicy), nullable=True))
    # Now migrate the data.  Don't import the table definition from the
    # models, it may break this migration when the model is updated in the
    # future (see the Alembic doc).
    mlist = sa.sql.table(
        'mailinglist',
        sa.sql.column('subscription_policy', Enum(SubscriptionPolicy))
        )
    # There were no enforced subscription policy before, so all lists are
    # considered open.
    op.execute(mlist.update().values(
        {'subscription_policy': op.inline_literal(SubscriptionPolicy.open)}))


def downgrade():
    if not is_sqlite(op.get_bind()):
        # SQLite does not support dropping columns.
        op.drop_column('mailinglist', 'subscription_policy')



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the NNTP runner and related utilities."""

import socket
import nntplib
import unittest

from email import message_from_bytes, utils
from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.interfaces.nntp import NewsgroupModeration
from mailman.runners import nntp
from mailman.testing.helpers import (
    configuration,
    get_queue_messages,
    LogFileMark,
    make_testable_runner,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from unittest import mock


class MockRet():
    """A mock to return a specific error."""
    def __init__(self):
        self.returncode = 2
        self.stderr = 'A mocked error message'


def mock_run(*args, **kwargs):
    return MockRet()


class TestPrepareMessage(unittest.TestCase):
    """Test message preparation."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._mlist.linked_newsgroup = 'example.test'
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A newsgroup posting
Message-ID: <ant>

Testing
""")

    def test_moderated_approved_header(self):
        # When the mailing list is moderated , the message will get an
        # Approved header, which NNTP software uses to forward to the
        # newsgroup.  The message would not have gotten to the mailing list if
        # it wasn't already approved.
        self._mlist.newsgroup_moderation = NewsgroupModeration.moderated
        nntp.prepare_message(self._mlist, self._msg, {})
        self.assertEqual(self._msg['approved'], 'test@example.com')

    def test_open_moderated_approved_header(self):
        # When the mailing list is moderated using an open posting policy, the
        # message will get an Approved header, which NNTP software uses to
        # forward to the newsgroup.  The message would not have gotten to the
        # mailing list if it wasn't already approved.
        self._mlist.newsgroup_moderation = NewsgroupModeration.open_moderated
        nntp.prepare_message(self._mlist, self._msg, {})
        self.assertEqual(self._msg['approved'], 'test@example.com')

    def test_moderation_removes_previous_approved_header(self):
        # Any existing Approved header is removed from moderated messages.
        self._msg['Approved'] = 'a bogus approval'
        self._mlist.newsgroup_moderation = NewsgroupModeration.moderated
        nntp.prepare_message(self._mlist, self._msg, {})
        headers = self._msg.get_all('approved')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'test@example.com')

    def test_open_moderation_removes_previous_approved_header(self):
        # Any existing Approved header is removed from moderated messages.
        self._msg['Approved'] = 'a bogus approval'
        self._mlist.newsgroup_moderation = NewsgroupModeration.open_moderated
        nntp.prepare_message(self._mlist, self._msg, {})
        headers = self._msg.get_all('approved')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'test@example.com')

    def test_stripped_subject(self):
        # The cook-headers handler adds the original and/or stripped (of the
        # prefix) subject to the metadata.  Assume that handler's been run;
        # check the Subject header.
        self._mlist.nntp_prefix_subject_too = False
        del self._msg['subject']
        self._msg['subject'] = 'Re: Your test'
        msgdata = dict(stripped_subject='Your test')
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        headers = self._msg.get_all('subject')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'Your test')

    def test_original_subject(self):
        # The cook-headers handler adds the original and/or stripped (of the
        # prefix) subject to the metadata.  Assume that handler's been run;
        # check the Subject header.
        self._mlist.nntp_prefix_subject_too = False
        del self._msg['subject']
        self._msg['subject'] = 'Re: Your test'
        msgdata = dict(original_subject='Your test')
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        headers = self._msg.get_all('subject')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'Your test')

    def test_stripped_subject_prefix_okay(self):
        # The cook-headers handler adds the original and/or stripped (of the
        # prefix) subject to the metadata.  Assume that handler's been run;
        # check the Subject header.
        self._mlist.nntp_prefix_subject_too = True
        del self._msg['subject']
        self._msg['subject'] = 'Re: Your test'
        msgdata = dict(stripped_subject='Your test')
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        headers = self._msg.get_all('subject')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'Re: Your test')

    def test_original_subject_prefix_okay(self):
        # The cook-headers handler adds the original and/or stripped (of the
        # prefix) subject to the metadata.  Assume that handler's been run;
        # check the Subject header.
        self._mlist.nntp_prefix_subject_too = True
        del self._msg['subject']
        self._msg['subject'] = 'Re: Your test'
        msgdata = dict(original_subject='Your test')
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        headers = self._msg.get_all('subject')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'Re: Your test')

    def test_add_newsgroups_header(self):
        # Prepared messages get a Newsgroups header.
        msgdata = dict(original_subject='Your test')
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        self.assertEqual(self._msg['newsgroups'], 'example.test')

    def test_add_newsgroups_header_to_existing(self):
        # If the message already has a Newsgroups header, the linked newsgroup
        # gets appended to that value, using comma-space separated lists.
        self._msg['Newsgroups'] = 'foo.test, bar.test'
        msgdata = dict(original_subject='Your test')
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        headers = self._msg.get_all('newsgroups')
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0], 'foo.test, bar.test, example.test')

    def test_add_lines_header(self):
        # A Lines: header seems useful.
        nntp.prepare_message(self._mlist, self._msg, {})
        self.assertEqual(self._msg['lines'], '1')

    def test_the_message_has_been_prepared(self):
        # A key gets added to the metadata so that a retry won't try to
        # re-apply all the preparations.
        msgdata = {}
        nntp.prepare_message(self._mlist, self._msg, msgdata)
        self.assertTrue(msgdata.get('prepped'))

    @configuration('nntp', remove_headers='x-complaints-to')
    def test_remove_headers(self):
        # During preparation, headers which cause problems with certain NNTP
        # servers such as INN get removed.
        self._msg['X-Complaints-To'] = 'arguments@example.com'
        nntp.prepare_message(self._mlist, self._msg, {})
        self.assertEqual(self._msg['x-complaints-to'], None)

    @configuration('nntp', rewrite_duplicate_headers="""
        To X-Original-To
        X-Fake X-Original-Fake
        """)
    def test_rewrite_headers(self):
        # Some NNTP servers are very strict about duplicate headers.  What we
        # can do is look at some headers and if they is more than one of that
        # header in the message, all the headers are deleted except the first
        # one, and then the other values are moved to the destination header.
        #
        # In this example, we'll create multiple To headers, which will all
        # get moved to X-Original-To.  However, because there will only be one
        # X-Fake header, it doesn't get rewritten.
        self._msg['To'] = 'test@example.org'
        self._msg['To'] = 'test@example.net'
        self._msg['X-Fake'] = 'ignore me'
        self.assertEqual(len(self._msg.get_all('to')), 3)
        self.assertEqual(len(self._msg.get_all('x-fake')), 1)
        nntp.prepare_message(self._mlist, self._msg, {})
        tos = self._msg.get_all('to')
        self.assertEqual(len(tos), 1)
        self.assertEqual(tos[0], 'test@example.com')
        original_tos = self._msg.get_all('x-original-to')
        self.assertEqual(len(original_tos), 2)
        self.assertEqual(original_tos,
                         ['test@example.org', 'test@example.net'])
        fakes = self._msg.get_all('x-fake')
        self.assertEqual(len(fakes), 1)
        self.assertEqual(fakes[0], 'ignore me')
        self.assertEqual(self._msg.get_all('x-original-fake'), None)

    @configuration('nntp', rewrite_duplicate_headers="""
        To X-Original-To
        X-Fake
        """)
    def test_odd_duplicates(self):
        # This is just a corner case, where there is an odd number of rewrite
        # headers.  In that case, the odd-one-out does not get rewritten.
        self._msg['x-fake'] = 'one'
        self._msg['x-fake'] = 'two'
        self._msg['x-fake'] = 'three'
        self.assertEqual(len(self._msg.get_all('x-fake')), 3)
        nntp.prepare_message(self._mlist, self._msg, {})
        fakes = self._msg.get_all('x-fake')
        self.assertEqual(len(fakes), 3)
        self.assertEqual(fakes, ['one', 'two', 'three'])


class TestNNTPRunner(unittest.TestCase):
    """The NNTP runner hands messages off to the NNTP server."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._mlist.linked_newsgroup = 'example.test'
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A newsgroup posting
Message-ID: <ant>

Testing
""")
        self._runner = make_testable_runner(nntp.NNTPRunner, 'nntp')
        self._nntpq = config.switchboards['nntp']

    @mock.patch('nntplib.NNTP')
    def test_connect(self, class_mock):
        # Test connection to the NNTP server with default values.
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        class_mock.assert_called_once_with(
            '', 119, user='', password='', readermode=True)

    @configuration('nntp', user='alpha', password='beta',
                   host='nntp.example.com', port='2112')
    @mock.patch('nntplib.NNTP')
    def test_connect_with_configuration(self, class_mock):
        # Test connection to the NNTP server with specific values.
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        class_mock.assert_called_once_with(
            'nntp.example.com', 2112,
            user='alpha', password='beta', readermode=True)

    @mock.patch('nntplib.NNTP')
    def test_post(self, class_mock):
        # Test that the message is posted to the NNTP server.
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        # Get the mocked instance, which was used in the runner.
        conn_mock = class_mock()
        # The connection object's post() method was called once with a
        # file-like object containing the message's bytes.  Read those bytes
        # and make some simple checks that the message is what we expected.
        args = conn_mock.post.call_args
        # One positional argument.
        self.assertEqual(len(args[0]), 1)
        # No keyword arguments.
        self.assertEqual(len(args[1]), 0)
        msg = message_from_bytes(args[0][0].read())
        self.assertEqual(msg['subject'], 'A newsgroup posting')

    @mock.patch('nntplib.NNTP')
    def test_post_long_message_id_not_folded(self, class_mock):
        # Test that the message with a long message-id is posted to the NNTP
        # server without folding.
        del self._msg['message-id']
        mid = utils.make_msgid('a_long_string_to_make_a_very_long_message-id',
                               self._mlist.mail_host)
        self._msg['Message-ID'] = mid
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        # Get the mocked instance, which was used in the runner.
        conn_mock = class_mock()
        # The connection object's post() method was called once with a
        # file-like object containing the message's bytes.  Read those bytes
        # and make some simple checks that the message is what we expected.
        args = conn_mock.post.call_args
        # One positional argument.
        self.assertEqual(len(args[0]), 1)
        # No keyword arguments.
        self.assertEqual(len(args[1]), 0)
        msg_bytes = args[0][0].read()
        msg = message_from_bytes(msg_bytes)
        self.assertEqual(msg['subject'], 'A newsgroup posting')
        self.assertIn(b'Message-ID: ' + mid.encode('us-ascii'), msg_bytes)

    @mock.patch('nntplib.NNTP')
    def test_connection_got_quit(self, class_mock):
        # The NNTP connection gets closed after a successful post.
        # Test that the message is posted to the NNTP server.
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        # Get the mocked instance, which was used in the runner.
        conn_mock = class_mock()
        # The connection object's post() method was called once with a
        # file-like object containing the message's bytes.  Read those bytes
        # and make some simple checks that the message is what we expected.
        conn_mock.quit.assert_called_once_with()

    @mock.patch('nntplib.NNTP', side_effect=nntplib.NNTPTemporaryError)
    def test_connect_with_nntplib_failure(self, class_mock):
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        mark = LogFileMark('mailman.error')
        self._runner.run()
        self.assertIn('NNTP error for test@example.com', mark.read())

    @mock.patch('nntplib.NNTP', side_effect=socket.error)
    def test_connect_with_socket_failure(self, class_mock):
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        mark = LogFileMark('mailman.error')
        self._runner.run()
        log_message = mark.readline()[:-1]
        self.assertTrue(log_message.endswith(
            'NNTP socket error for test@example.com'))

    @mock.patch('nntplib.NNTP', side_effect=RuntimeError)
    def test_connect_with_other_failure(self, class_mock):
        # In this failure mode, the message stays queued, so we can only run
        # the nntp runner once.
        def once(runner):
            # I.e. stop immediately, since the queue will not be empty.
            return True
        runner = make_testable_runner(nntp.NNTPRunner, 'nntp', predicate=once)
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        mark = LogFileMark('mailman.error')
        runner.run()
        log_message = mark.readline()[:-1]
        self.assertTrue(log_message.endswith(
            'NNTP unexpected exception for test@example.com'))
        items = get_queue_messages('nntp', expected_count=1)
        self.assertEqual(items[0].msgdata['listid'], 'test.example.com')
        self.assertEqual(items[0].msg['subject'], 'A newsgroup posting')

    @mock.patch('nntplib.NNTP', side_effect=nntplib.NNTPTemporaryError)
    def test_connection_never_gets_quit_after_failures(self, class_mock):
        # The NNTP connection doesn't get closed after an unsuccessful
        # connection, since there's nothing to close.
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        # Get the mocked instance, which was used in the runner.  Turn off the
        # exception raising side effect first though!
        class_mock.side_effect = None
        conn_mock = class_mock()
        # The connection object's post() method was called once with a
        # file-like object containing the message's bytes.  Read those bytes
        # and make some simple checks that the message is what we expected.
        self.assertEqual(conn_mock.quit.call_count, 0)

    @mock.patch('nntplib.NNTP')
    def test_connection_got_quit_after_post_failure(self, class_mock):
        # The NNTP connection does get closed after an unsuccessful post.
        # Add a side-effect to the instance mock's .post() method.
        conn_mock = class_mock()
        conn_mock.post.side_effect = nntplib.NNTPTemporaryError
        # Set the Message-ID to ours so the message isn't retried.
        del self._msg['message-id']
        self._msg['message-id'] = utils.make_msgid(self._mlist.list_name,
                                                   self._mlist.mail_host)
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        self._runner.run()
        # The connection object's post() method was called once with a
        # file-like object containing the message's bytes.  The
        # NNTPTemporaryError will not be retried because the Message-ID is
        # ours.
        conn_mock.quit.assert_called_once_with()

    @mock.patch('nntplib.NNTP')
    def test_connection_got_two_quit_after_post_failure(self, class_mock):
        # The NNTP connection does get closed after an unsuccessful post.
        # Add a side-effect to the instance mock's .post() method.
        conn_mock = class_mock()
        conn_mock.post.side_effect = nntplib.NNTPTemporaryError
        self._nntpq.enqueue(self._msg, {}, listid='test.example.com')
        mark = LogFileMark('mailman.error')
        self._runner.run()
        # The connection object's post() method was called twice with a
        # file-like object containing the message's bytes.  The
        # NNTPTemporaryError is retried with a munged Message-ID because it
        # might have been due to a duplicate Message-ID.  Check that quit is
        # called twice.
        self.assertEqual(conn_mock.quit.call_count, 2)
        # There should be a log message with the original Message-ID.
        self.assertIn('<ant> NNTP error for test@example.com', mark.read())

    @mock.patch('subprocess.run', mock_run)
    def test_do_periodic_invokes_mailman_gatenews(self):
        # _do_periodic should invoke mailman gatenews which will fail.
        mark = LogFileMark('mailman.error')
        nntp.NNTPRunner('nntp')._do_periodic()
        log_message = mark.read()
        self.assertIn('Running nntp runner periodic task gatenews',
                      log_message)
        self.assertIn('gatenews failed. status: 2', log_message)
        self.assertIn('message: A mocked error message', log_message)



# Copyright (C) 2001-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Calculate the list owner recipients (includes moderators)."""

from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.handler import IHandler
from mailman.interfaces.member import DeliveryStatus
from public import public
from zope.interface import implementer


@public
@implementer(IHandler)
class OwnerRecipients:
    """Calculate the owner (and moderator) recipients for -owner postings."""

    name = 'owner-recipients'
    description = _('Calculate the owner and moderator recipients.')

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        # Short circuit if we've already calculated the recipients list,
        # regardless of whether the list is empty or not.
        if 'recipients' in msgdata:
            return
        # -owner messages go to both the owners and moderators, which is most
        # conveniently accessed via the administrators roster.
        recipients = set(admin.address.email
                         for admin in mlist.administrators.members
                         if admin.delivery_status == DeliveryStatus.enabled)
        # To prevent -owner messages from going into a black hole, if there
        # are no administrators available, the message goes to the site owner.
        if len(recipients) == 0:
            # Ensure that the site owner address is a unicode.
            # See LP: #1130957
            site_owner = config.mailman.site_owner
            if isinstance(site_owner, bytes):
                site_owner = site_owner.decode('utf-8')
            msgdata['recipients'] = set((site_owner,))
        else:
            msgdata['recipients'] = recipients
        # Don't decorate these messages with the header/footers.  Eventually
        # we should support unique decorations for owner emails.
        msgdata['nodecorate'] = True
        # We should probably always VERP deliveries to the owners.  We
        # *really* want to know if they are bouncing.
        msgdata['verp'] = True



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Application support for pipeline processing."""

import logging

from mailman.app.bounces import bounce_message
from mailman.config import config
from mailman.interfaces.handler import IHandler
from mailman.interfaces.pipeline import (
    DiscardMessage,
    IPipeline,
    RejectMessage,
)
from mailman.utilities.modules import add_components
from public import public


dlog = logging.getLogger('mailman.debug')
vlog = logging.getLogger('mailman.vette')


@public
def process(mlist, msg, msgdata, pipeline_name='built-in'):
    """Process the message through the given pipeline.

    :param mlist: the IMailingList for this message.
    :param msg: The Message object.
    :param msgdata: The message metadata dictionary.
    :param pipeline_name: The name of the pipeline to process through.
    """
    message_id = msg.get('message-id', 'n/a')
    pipeline = config.pipelines[pipeline_name]
    for handler in pipeline:
        dlog.debug('{} pipeline {} processing: {}'.format(
            message_id, pipeline_name, handler.name))
        try:
            handler.process(mlist, msg, msgdata)
        except DiscardMessage as error:
            vlog.info(
                '{} discarded by "{}" pipeline handler "{}": {}'.format(
                    message_id, pipeline_name, handler.name, error.message))
            # Stop processing the pipeline.
            break
        except RejectMessage as error:
            vlog.info(
                '{} rejected by "{}" pipeline handler "{}": {}'.format(
                    message_id, pipeline_name, handler.name, str(error)))
            bounce_message(mlist, msg, error)
            # Stop processing the pipeline.
            break


@public
def initialize():
    """Initialize the pipelines."""
    # Find all handlers in the registered plugins.
    add_components('handlers', IHandler, config.handlers)
    # Set up some pipelines.
    add_components('pipelines', IPipeline, config.pipelines)



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test email address registration."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.mailinglist import SubscriptionPolicy
from mailman.interfaces.member import MemberRole
from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import ISubscriptionManager, TokenOwner
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import get_queue_messages
from mailman.testing.layers import ConfigLayer
from mailman.utilities.datetime import now
from zope.component import getUtility


class TestRegistrar(unittest.TestCase):
    """Test registration."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._registrar = ISubscriptionManager(self._mlist)
        self._pendings = getUtility(IPendings)
        self._anne = getUtility(IUserManager).create_address(
            'anne@example.com')

    def test_initial_conditions(self):
        # Registering a subscription request provides a unique token associated
        # with a pendable, and the owner of the token.
        self.assertEqual(self._pendings.count(), 0)
        token, token_owner, member = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.subscriber)
        self.assertIsNone(member)
        self.assertEqual(self._pendings.count(), 1)
        record = self._pendings.confirm(token, expunge=False)
        self.assertEqual(record['list_id'], self._mlist.list_id)
        self.assertEqual(record['email'], 'anne@example.com')

    def test_subscribe(self):
        # Registering a subscription request where no confirmation or
        # moderation steps are needed, leaves us with no token or owner, since
        # there's nothing more to do.
        self._mlist.subscription_policy = SubscriptionPolicy.open
        self._anne.verified_on = now()
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertEqual(rmember, member)
        self.assertEqual(member.address, self._anne)
        # There's nothing to confirm.
        record = self._pendings.confirm(token, expunge=False)
        self.assertIsNone(record)

    def test_no_such_token(self):
        # Given a token which is not in the database, a LookupError is raised.
        self._registrar.register(self._anne)
        self.assertRaises(LookupError, self._registrar.confirm, 'not-a-token')

    def test_confirm_because_verify(self):
        # We have a subscription request which requires the user to confirm
        # (because she does not have a verified address), but not the moderator
        # to approve.  Running the workflow gives us a token.  Confirming the
        # token subscribes the user.
        self._mlist.subscription_policy = SubscriptionPolicy.open
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.subscriber)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Now confirm the subscription.
        token, token_owner, rmember = self._registrar.confirm(token)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertEqual(rmember, member)
        self.assertEqual(member.address, self._anne)

    def test_confirm_because_confirm(self):
        # We have a subscription request which requires the user to confirm
        # (because of list policy), but not the moderator to approve.  Running
        # the workflow gives us a token.  Confirming the token subscribes the
        # user.
        self._mlist.subscription_policy = SubscriptionPolicy.confirm
        self._anne.verified_on = now()
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.subscriber)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Now confirm the subscription.
        token, token_owner, rmember = self._registrar.confirm(token)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertEqual(rmember, member)
        self.assertEqual(member.address, self._anne)

    def test_confirm_because_moderation(self):
        # We have a subscription request which requires the moderator to
        # approve.  Running the workflow gives us a token.  Confirming the
        # token subscribes the user.
        self._mlist.subscription_policy = SubscriptionPolicy.moderate
        self._anne.verified_on = now()
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.moderator)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Now confirm the subscription.
        token, token_owner, rmember = self._registrar.confirm(token)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertEqual(rmember, member)
        self.assertEqual(member.address, self._anne)

    def test_confirm_because_confirm_then_moderation(self):
        # We have a subscription request which requires the user to confirm
        # (because she does not have a verified address) and the moderator to
        # approve.  Running the workflow gives us a token.  Confirming the
        # token runs the workflow a little farther, but still gives us a
        # token.  Confirming again subscribes the user.
        self._mlist.subscription_policy = (
            SubscriptionPolicy.confirm_then_moderate)
        self._anne.verified_on = now()
        # Runs until subscription confirmation.
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.subscriber)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Now confirm the subscription, and wait for the moderator to approve
        # the subscription.  She is still not subscribed.
        new_token, token_owner, rmember = self._registrar.confirm(token)
        # The new token, used for the moderator to approve the message, is not
        # the same as the old token.
        self.assertNotEqual(new_token, token)
        self.assertIsNotNone(new_token)
        self.assertEqual(token_owner, TokenOwner.moderator)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Confirm once more, this time as the moderator approving the
        # subscription.  Now she's a member.
        token, token_owner, rmember = self._registrar.confirm(new_token)
        self.assertIsNone(token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertEqual(rmember, member)
        self.assertEqual(member.address, self._anne)

    def test_confirm_then_moderate_with_different_tokens(self):
        # Ensure that the confirmation token the user sees when they have to
        # confirm their subscription is different than the token the moderator
        # sees when they approve the subscription.  This prevents the user
        # from using a replay attack to subvert moderator approval.
        self._mlist.subscription_policy = (
            SubscriptionPolicy.confirm_then_moderate)
        self._anne.verified_on = now()
        # Runs until subscription confirmation.
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.subscriber)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Now confirm the subscription, and wait for the moderator to approve
        # the subscription.  She is still not subscribed.
        new_token, token_owner, rmember = self._registrar.confirm(token)
        # The status is not true because the user has not yet been subscribed
        # to the mailing list.
        self.assertIsNotNone(new_token)
        self.assertEqual(token_owner, TokenOwner.moderator)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # The new token is different than the old token.
        self.assertNotEqual(token, new_token)
        # Trying to confirm with the old token does not work.
        self.assertRaises(LookupError, self._registrar.confirm, token)
        # Confirm once more, this time with the new token, as the moderator
        # approving the subscription.  Now she's a member.
        done_token, token_owner, rmember = self._registrar.confirm(new_token)
        # The token is None, signifying that the member has been subscribed.
        self.assertIsNone(done_token)
        self.assertEqual(token_owner, TokenOwner.no_one)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertEqual(rmember, member)
        self.assertEqual(member.address, self._anne)

    def test_discard_waiting_for_confirmation(self):
        # While waiting for a user to confirm their subscription, we discard
        # the workflow.
        self._mlist.subscription_policy = SubscriptionPolicy.confirm
        self._anne.verified_on = now()
        # Runs until subscription confirmation.
        token, token_owner, rmember = self._registrar.register(self._anne)
        self.assertIsNotNone(token)
        self.assertEqual(token_owner, TokenOwner.subscriber)
        self.assertIsNone(rmember)
        member = self._mlist.regular_members.get_member('anne@example.com')
        self.assertIsNone(member)
        # Now discard the subscription request.
        self._registrar.discard(token)
        # Trying to confirm the token now results in an exception.
        self.assertRaises(LookupError, self._registrar.confirm, token)

    def test_admin_notify_mchanges(self):
        # When a user gets subscribed via the subscription policy workflow,
        # the list administrators get an email notification.
        self._mlist.subscription_policy = SubscriptionPolicy.open
        self._mlist.admin_notify_mchanges = True
        self._mlist.send_welcome_message = False
        token, token_owner, member = self._registrar.register(
            self._anne, pre_verified=True)
        # Anne is now a member.
        self.assertEqual(member.address.email, 'anne@example.com')
        # And there's a notification email waiting for Bart.
        items = get_queue_messages('virgin', expected_count=1)
        message = items[0].msg
        self.assertEqual(message['To'], 'ant-owner@example.com')
        self.assertEqual(message['Subject'], 'Ant subscription notification')
        self.assertEqual(message.get_payload(), """\
anne@example.com has been successfully subscribed to Ant.
""")

    def test_no_admin_notify_mchanges(self):
        # Even when a user gets subscribed via the subscription policy
        # workflow, the list administrators won't get an email notification if
        # they don't want one.
        self._mlist.subscription_policy = SubscriptionPolicy.open
        self._mlist.admin_notify_mchanges = False
        self._mlist.send_welcome_message = False
        # Bart is an administrator of the mailing list.
        bart = getUtility(IUserManager).create_address(
            'bart@example.com', 'Bart Person')
        self._mlist.subscribe(bart, MemberRole.owner)
        token, token_owner, member = self._registrar.register(
            self._anne, pre_verified=True)
        # Anne is now a member.
        self.assertEqual(member.address.email, 'anne@example.com')
        # There's no notification email waiting for Bart.
        get_queue_messages('virgin', expected_count=0)



# Copyright (C) 1998-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Standard Mailman message object.

This is a subclass of email.message.Message but provides a slightly extended
interface which is more convenient for use inside Mailman.  It also supports
safe pickle deserialization, even if the email package adds additional Message
attributes.
"""

import re
import email
import email.utils
import email.message

from email.header import decode_header, Header, make_header
from email.mime.multipart import MIMEMultipart
from mailman.config import config
from mailman.interfaces.address import IEmailValidator
from public import public
from zope.component import getUtility


COMMASPACE = ', '


@public
class Message(email.message.Message):
    # BAW: For debugging w/ bin/dumpdb.  Apparently pprint uses repr.
    def __repr__(self):
        return self.__str__()

    def __setstate__(self, values):
        self.__dict__ = values

    def as_string(self):
        # Work around for https://bugs.python.org/issue27321 and
        # https://bugs.python.org/issue32330.
        try:
            value = email.message.Message.as_string(self)
        except (KeyError, LookupError, UnicodeEncodeError):
            value = email.message.Message.as_bytes(self).decode(
                'ascii', 'replace')
        # Also ensure no unicode surrogates in the returned string.
        return email.utils._sanitize(value)

    def as_bytes(self):
        # Workaround for https://bugs.python.org/issue41307.
        try:
            value = email.message.Message.as_bytes(self)
        except UnicodeEncodeError:
            value = email.message.Message.as_string(self).encode('utf-8')
        return value

    @property
    def sender(self):
        """The address considered to be the author of the email.

        This is the first non-None value in the list of senders.

        :return: The email address of the first found sender, or the empty
            string if no sender address was found.
        :rtype: email address
        """
        for address in self.senders:
            # This could be None or the empty string.
            if address:
                return address
        return ''

    @property
    def senders(self):
        """Return a list of addresses representing the author of the email.

        The list will contain email addresses in the order determined by the
        configuration variable `sender_headers` in the `[mailman]` section.
        By default it uses this list of headers in order:

        1. From:
        2. envelope sender (i.e. From_, unixfrom, or RFC 2821 MAIL FROM)
        3. Reply-To:
        4. Sender:

        The return addresses are guaranteed to be lower case or None.  There
        may be more than four values in the returned list, since some of the
        originator headers above can appear multiple times in the message, or
        contain multiple values.

        :return: The list of email addresses that can be considered the sender
            of the message.
        :rtype: A list of email addresses or Nones
        """
        envelope_sender = self.get_unixfrom()
        senders = []
        for header in config.mailman.sender_headers.split():
            header = header.lower()
            if header == 'from_':
                senders.append(envelope_sender.lower()
                               if envelope_sender is not None
                               else '')
            else:
                for field_value in self.get_all(header, []):
                    # Convert the header to str in case it's a Header instance.
                    header_value = re.sub(
                        '[\r\n]',
                        '',
                        str(make_header(decode_header(field_value)))
                        )
                    for name, address in email.utils.getaddresses(
                            [header_value]):
                        senders.append(address.lower())
        # Filter out invalid addresses, None and the empty string, and convert
        # to unicode.
        clean_senders = []
        validator = getUtility(IEmailValidator)
        for sender in senders:
            if not sender:
                continue
            if isinstance(sender, bytes):
                sender = sender.decode('ascii')
            if not validator.is_valid(sender):
                continue
            clean_senders.append(sender)
        return clean_senders


@public
class MultipartDigestMessage(MIMEMultipart, Message):
    """Mix-in class for MIME digest messages."""


@public
class UserNotification(Message):
    """Class for internally crafted messages."""

    def __init__(self, recipients, sender, subject=None, text=None, lang=None):
        Message.__init__(self)
        charset = (lang.charset if lang is not None else 'us-ascii')
        subject = ('(no subject)' if subject is None else subject)
        if text is not None:
            try:
                self.set_payload(text.encode(charset), charset)
            except UnicodeError:
                self.set_payload(text.encode(), 'utf-8')
        self['Subject'] = Header(
            subject, charset, header_name='Subject', errors='replace')
        self['From'] = sender
        if isinstance(recipients, (list, set, tuple)):
            self['To'] = COMMASPACE.join(recipients)
            self.recipients = {email.utils.parseaddr(recipient)[1] for
                               recipient in recipients}
        else:
            self['To'] = recipients
            self.recipients = set([email.utils.parseaddr(recipients)[1]])

    def send(self, mlist, *, add_precedence=True, **_kws):
        """Sends the message by enqueuing it to the 'virgin' queue.

        This is used for all internally crafted messages.

        :param mlist: The mailing list to send the message to.
        :type mlist: `IMailingList`
        :param add_precedence: Flag indicating whether a `Precedence: bulk`
            header should be added to the message or not.
        :type add_precedence: bool

        This function also accepts arbitrary keyword arguments.  The key/value
        pairs for **kws is added to the metadata dictionary associated with
        the enqueued message.
        """
        # Since we're crafting the message from whole cloth, let's make sure
        # this message has a Message-ID.
        if 'message-id' not in self:
            self['Message-ID'] = email.utils.make_msgid()
        # Ditto for Date: as required by RFC 2822.
        if 'date' not in self:
            self['Date'] = email.utils.formatdate(localtime=True)
        # UserNotifications are typically for admin messages, and for messages
        # other than list explosions.  Send these out as Precedence: bulk, but
        # don't override an existing Precedence: header.
        if 'precedence' not in self and add_precedence:
            self['Precedence'] = 'bulk'
        self._enqueue(mlist, **_kws)

    def _enqueue(self, mlist, **_kws):
        # Not imported at module scope to avoid import loop
        virginq = config.switchboards['virgin']
        # The message metadata better have a 'recip' attribute.
        enqueue_kws = dict(
            recipients=self.recipients,
            nodecorate=True,
            reduced_list_headers=True,
            )
        if mlist is not None:
            enqueue_kws['listid'] = mlist.list_id
        enqueue_kws.update(_kws)
        virginq.enqueue(self, **enqueue_kws)


@public
class OwnerNotification(UserNotification):
    """Like user notifications, but this message goes to some owners."""

    def __init__(self, mlist, subject=None, text=None, roster=None):
        if roster is None:
            recipients = set([config.mailman.site_owner])
            to = config.mailman.site_owner
        else:
            recipients = set(address.email for address in roster.addresses)
            to = mlist.owner_address
        sender = config.mailman.site_owner
        UserNotification.__init__(self, recipients, sender, subject,
                                  text, mlist.preferred_language)
        # Hack the To header to look like it's going to the -owner address
        del self['to']
        self['To'] = to
        self._sender = sender

    def _enqueue(self, mlist, **_kws):
        # Not imported at module scope to avoid import loop
        virginq = config.switchboards['virgin']
        # The message metadata better have a `recip' attribute
        virginq.enqueue(self,
                        listid=mlist.list_id,
                        recipients=self.recipients,
                        nodecorate=True,
                        reduced_list_headers=True,
                        envsender=self._sender,
                        **_kws)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `banned-address` rule."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.bans import IBanManager
from mailman.interfaces.usermanager import IUserManager
from mailman.rules import banned_address
from mailman.testing.helpers import (
    set_preferred,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestBannedAddress(unittest.TestCase):
    """Test the banned-address rule."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')

    def test_no_banned_sender(self):
        # Simple case where the sender is not banned.
        user_manager = getUtility(IUserManager)
        anne = user_manager.create_user('anne@example.com')
        set_preferred(anne)
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A test message
Message-ID: <ant>
MIME-Version: 1.0

A message body.
""")
        rule = banned_address.BannedAddress()
        result = rule.check(self._mlist, msg, {})
        self.assertFalse(result)

    def test_simple_banned_sender(self):
        # Simple case where the sender is banned.
        user_manager = getUtility(IUserManager)
        anne = user_manager.create_user('anne@example.com')
        set_preferred(anne)
        IBanManager(self._mlist).ban('anne@example.com')
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A test message
Message-ID: <ant>
MIME-Version: 1.0

A message body.
""")
        rule = banned_address.BannedAddress()
        result = rule.check(self._mlist, msg, {})
        self.assertTrue(result)

    def test_rule_returns_reason(self):
        # Ensure a reason is returned.
        user_manager = getUtility(IUserManager)
        anne = user_manager.create_user('anne@example.com')
        set_preferred(anne)
        IBanManager(self._mlist).ban('anne@example.com')
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A test message
Message-ID: <ant>
MIME-Version: 1.0

A message body.
""")
        rule = banned_address.BannedAddress()
        msgdata = {}
        result = rule.check(self._mlist, msg, msgdata)
        self.assertTrue(result)
        self.assertEqual(
            msgdata['moderation_reasons'],
            [('Message sender {} is banned from this list',
              'anne@example.com')])

    def test_banned_address_linked_to_user(self):
        # Anne is subscribed to a mailing list as a user with her preferred
        # address.  She also has a secondary address which is banned and which
        # she uses to post to the mailing list.  The rule matches because the
        # posting address is banned.
        user_manager = getUtility(IUserManager)
        anne = user_manager.create_user('anne@example.com')
        set_preferred(anne)
        anne.link(user_manager.create_address('anne.person@example.com'))
        IBanManager(self._mlist).ban('anne.person@example.com')
        msg = mfs("""\
From: anne.person@example.com
To: test@example.com
Subject: A test message
Message-ID: <ant>
MIME-Version: 1.0

A message body.
""")
        rule = banned_address.BannedAddress()
        result = rule.check(self._mlist, msg, {})
        self.assertTrue(result)

    def test_banned_sender_among_multiple_senders(self):
        # Two addresses are created, one of which is banned.  The rule matches
        # because all senders are checked.
        user_manager = getUtility(IUserManager)
        user_manager.create_address('anne@example.com')
        user_manager.create_address('bart@example.com')
        IBanManager(self._mlist).ban('bart@example.com')
        msg = mfs("""\
From: anne@example.com
Sender: bart@example.com
To: test@example.com
Subject: A test message
Message-ID: <ant>
MIME-Version: 1.0

A message body.
""")
        rule = banned_address.BannedAddress()
        result = rule.check(self._mlist, msg, {})
        self.assertTrue(result)



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Additional tests for the `lists` command line subcommand."""

import unittest

from click.testing import CliRunner
from mailman.app.lifecycle import create_list
from mailman.commands.cli_lists import lists
from mailman.interfaces.domain import IDomainManager
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestLists(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._command = CliRunner()

    def test_lists_with_domain_option(self):
        # LP: #1166911 - non-matching lists were returned.
        getUtility(IDomainManager).add(
            'example.net', 'An example domain.')
        create_list('test1@example.com')
        create_list('test2@example.com')
        # Only this one should show up.
        create_list('test3@example.net')
        create_list('test4@example.com')
        result = self._command.invoke(lists, ('--domain', 'example.net'))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(
            result.output,
            '1 matching mailing lists found:\ntest3@example.net\n')



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the admin-notify subcommand."""

import unittest

from click.testing import CliRunner
from mailman.app.lifecycle import create_list
from mailman.app.moderator import hold_message
from mailman.chains.hold import HoldChain
from mailman.commands.cli_notify import notify
from mailman.interfaces.mailinglist import SubscriptionPolicy
from mailman.interfaces.subscriptions import ISubscriptionManager
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import (
    get_queue_messages,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestNotifyCommand(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._mlist2 = create_list('bee@example.com')
        self._mlist.subscription_policy = SubscriptionPolicy.moderate
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        msg = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: message 1

""")
        # Hold this message.
        hold_message(self._mlist, msg, {}, 'Non-member post')
        # And a second one too.
        msg2 = mfs("""\
To: ant@example.com
From: bob@example.com
Subject: =?utf-8?q?message_2?=

""")
        hold_message(self._mlist, msg2, {}, 'Some other reason')
        usermanager = getUtility(IUserManager)
        submanager = ISubscriptionManager(self._mlist)
        # Generate held subscription.
        usera = usermanager.make_user('anne@example.com')
        usera.addresses[0].verified_on = usera.addresses[0].registered_on
        usera.preferred_address = usera.addresses[0]
        submanager.register(usera)
        # Generate a held unsubscription.
        userb = usermanager.make_user('bob@example.com')
        userb.addresses[0].verified_on = userb.addresses[0].registered_on
        userb.preferred_address = userb.addresses[0]
        submanager.register(userb, pre_verified=True, pre_confirmed=True,
                            pre_approved=True)
        submanager.unregister(userb)
        self._command = CliRunner()

    def test_notify_dry_run_verbose(self):
        # Clear messages from setup.
        get_queue_messages('virgin')
        result = self._command.invoke(notify,
                                      ('-n', '-v', '-l', 'ant@example.com'))
        self.assertMultiLineEqual(result.output, """\
The ant@example.com list has 4 moderation requests waiting.
""")
        get_queue_messages('virgin', expected_count=0)

    def test_notify_dry_run_verbose_list_id(self):
        # Clear messages from setup.
        get_queue_messages('virgin')
        result = self._command.invoke(notify,
                                      ('-n', '-v', '-l', 'ant.example.com'))
        self.assertMultiLineEqual(result.output, """\
The ant@example.com list has 4 moderation requests waiting.
""")
        get_queue_messages('virgin', expected_count=0)

    def test_notify_one_list(self):
        # Clear messages from setup.
        get_queue_messages('virgin')
        result = self._command.invoke(notify, ('-v', '-l', 'ant@example.com'))
        self.assertMultiLineEqual(result.output, """\
The ant@example.com list has 4 moderation requests waiting.
""")
        msg = get_queue_messages('virgin', expected_count=1)[0].msg
        self.assertMultiLineEqual(msg.get_payload(), """\
The ant@example.com list has 4 moderation requests waiting.


Held Subscriptions:
    User: anne@example.com

Held Unsubscriptions:
    User: bob@example.com

Held Messages:
    Sender: anne@example.com
    Subject: message 1
    Reason: Non-member post

    Sender: bob@example.com
    Subject: message 2
    Reason: Some other reason


Please attend to this at your earliest convenience.""")

    def test_notify_bogus_list(self):
        result = self._command.invoke(notify,
                                      ('-v', '-l', 'bogus@example.com'))
        self.assertMultiLineEqual(result.output, """\
No such list found: bogus@example.com
""")

    def test_notify_all_lists(self):
        # Clear messages from setup.
        get_queue_messages('virgin')
        result = self._command.invoke(notify, ('-v',))
        self.assertMultiLineEqual(result.output, """\
The ant@example.com list has 4 moderation requests waiting.
The bee@example.com list has 0 moderation requests waiting.
""")
        msg = get_queue_messages('virgin', expected_count=1)[0].msg
        self.assertMultiLineEqual(msg.get_payload(), """\
The ant@example.com list has 4 moderation requests waiting.


Held Subscriptions:
    User: anne@example.com

Held Unsubscriptions:
    User: bob@example.com

Held Messages:
    Sender: anne@example.com
    Subject: message 1
    Reason: Non-member post

    Sender: bob@example.com
    Subject: message 2
    Reason: Some other reason


Please attend to this at your earliest convenience.""")


class TestBogusRFC2047(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        msg = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: =?gb2312?q?=E9=FF_message_1?=

""")
        # Hold this message.
        hold_message(self._mlist, msg, {}, 'Non-member post')
        self._command = CliRunner()

    def test_bogus_rfc2047(self):
        # Clear messages from setup.
        get_queue_messages('virgin')
        result = self._command.invoke(notify, ('-v',))
        self.assertMultiLineEqual(result.output, """\
The ant@example.com list has 1 moderation requests waiting.
""")
        msg = get_queue_messages('virgin', expected_count=1)[0].msg
        self.assertMultiLineEqual(msg.get_payload(), """\
The ant@example.com list has 1 moderation requests waiting.


Held Messages:
    Sender: anne@example.com
    Subject: =?gb2312?q?=E9=FF_message_1?=
    Reason: Non-member post


Please attend to this at your earliest convenience.""")


class TestChainsHold(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        msg = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: message 1

""")
        # Hold this message using chains.hold. Supply moderation_reasons and
        # set fromusenet to suppress user notice.
        msgdata = {'moderation_reasons': ['Non-member post'],
                   'fromusenet': True}
        # Suppress owner notice too.
        self._mlist.admin_immed_notify = False
        HoldChain()._process(self._mlist, msg, msgdata)
        self._command = CliRunner()

    def test_notify_with_hold_from_hold_chain(self):
        # Clear messages from setup.
        get_queue_messages('virgin')
        result = self._command.invoke(notify, ('-v',))
        self.assertMultiLineEqual(result.output, """\
The ant@example.com list has 1 moderation requests waiting.
""")
        msg = get_queue_messages('virgin', expected_count=1)[0].msg
        self.assertMultiLineEqual(msg.get_payload(), """\
The ant@example.com list has 1 moderation requests waiting.


Held Messages:
    Sender: anne@example.com
    Subject: message 1
    Reason: Non-member post


Please attend to this at your earliest convenience.""")



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Database type conversions."""

import uuid

from public import public
from sqlalchemy import Integer
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import CHAR, Text, TypeDecorator, Unicode


@public
class Enum(TypeDecorator):
    """Handle Python 3.4 style enums.

    Stores an integer-based Enum as an integer in the database, and
    converts it on-the-fly.
    """
    impl = Integer
    cache_ok = True

    def __init__(self, enum, *args, **kw):
        super().__init__(*args, **kw)
        self.enum = enum

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum(value)


@public
class UUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(postgresql.UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return '%.32x' % value.int

    def process_result_value(self, value, dialect):
        # postgres returns UUID object by default in SQLAlchemy 2.0.0
        # and str value on sqlalchemy < 2.0.0.
        if (value is None) or (isinstance(value, uuid.UUID)):
            return value
        else:
            return uuid.UUID(value)


@public
class SAUnicode(TypeDecorator):
    """Unicode datatype to support fixed length VARCHAR in MySQL.

    This type compiles to VARCHAR(255) in case of MySQL, and in case of
    other dialects defaults to the Unicode type.  This was created so
    that we don't have to alter the output of the default Unicode data
    type and it can still be used if needed in the codebase.

    WARNING: On MySQL it will not be possible to insert emojis or other
    4-byte utf-8 encoded characters. This should only be used if the entire
    column needs to be indexed, otherwise use SAUnicode4Byte.
    """
    impl = Unicode
    cache_ok = True


@compiles(SAUnicode)
def default_sa_unicode(element, compiler, **kw):
    return compiler.visit_unicode(element, **kw)


@compiles(SAUnicode, 'mysql')
def compile_sa_unicode(element, compiler, **kw):
    # We hardcode the collate here to make string comparison case sensitive.
    return 'VARCHAR(255) COLLATE utf8_bin'


@public
class SAUnicode4Byte(TypeDecorator):
    """Unicode datatype to support fixed length VARCHAR in MySQL.

    This type compiles to VARCHAR(255) in case of MySQL, and in case of
    other dialects defaults to the Unicode type.  This was created so
    that we don't have to alter the output of the default Unicode data
    type and it can still be used if needed in the codebase.
    """
    impl = Unicode
    cache_ok = True


@compiles(SAUnicode4Byte)
def default_sa_unicode_4byte(element, compiler, **kw):
    return compiler.visit_unicode(element, **kw)


@compiles(SAUnicode4Byte, 'mysql')
def compile_sa_unicode_4byte(element, compiler, **kw):  # pragma: nocover
    # We hardcode the collate here to make string comparison case sensitive.
    return 'VARCHAR(255) COLLATE utf8mb4_bin'


@public
class SAUnicodeLarge(TypeDecorator):
    """Similar to SAUnicode type, but compiles to VARCHAR(510).

    This is double size of SAUnicode defined above.
    """
    impl = Unicode
    cache_ok = True


@compiles(SAUnicodeLarge, 'mysql')
def compile_sa_unicode_large(element, compiler, **kw):  # pragma: nocover
    # We hardcode the collate here to make string comparison case sensitive.
    return 'VARCHAR(510) COLLATE utf8mb4_bin'


@compiles(SAUnicodeLarge)
def default_sa_unicode_large(element, compiler, **kw):
    return compiler.visit_unicode(element, **kw)


@public
class SAUnicodeXL(TypeDecorator):
    """Similar to SAUnicode type, but compiles to VARCHAR(20000).

    This allows very long values up to 20,000 bytes for cases where
    SAUnicodeLarge is not enough.

    Because of MySQL issues columns of this type should either not be indexed
    or should be indexed with an Index() function specifying mysql_length=N
    where N is (probably much) less than 767.
    See https://docs.sqlalchemy.org/en/latest/dialects/mysql.html#index-length
    """
    impl = Unicode
    cache_ok = True


@compiles(SAUnicodeXL, 'mysql')
def compile_sa_unicode_xl(element, compiler, **kw):
    # We hardcode the collate here to make string comparison case sensitive.
    return 'VARCHAR(20000) COLLATE utf8_bin'              # pragma: nocover


@compiles(SAUnicodeXL)
def default_sa_unicode_xl(element, compiler, **kw):
    return compiler.visit_unicode(element, **kw)


@public
class SAText(TypeDecorator):
    """Text datatype to support large text content in MySQL.

    This type compiles to TEXT COLLATE utf8mb4_bin in case of MySQL, and in
    case of other dialects defaults to the Text type.
    """
    impl = Text
    cache_ok = True


@compiles(SAText)
def default_sa_text(element, compiler, **kw):
    return compiler.visit_text(element, **kw)


@compiles(SAText, 'mysql')
def compile_sa_text(element, compiler, **kw):
    # We hardcode the collate here to make string comparison case sensitive.
    return 'TEXT COLLATE utf8mb4_bin'                     # pragma: nocover



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The `mailman shell` subcommand."""

import re
import sys
import click

from contextlib import ExitStack, suppress
from functools import partial
from lazr.config import as_boolean
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.command import ICLISubCommand
from mailman.interfaces.listmanager import IListManager
from mailman.utilities.interact import DEFAULT_BANNER, interact
from mailman.utilities.modules import call_name
from mailman.utilities.options import I18nCommand
from public import public
from string import Template
from traceback import print_exc
from zope.component import getUtility
from zope.interface import implementer


# Global holding onto the open mailing list.
m = None
# Global holding the results of --run.
r = None


def start_ipython1(overrides, banner, *, debug=False):
    try:
        from IPython.frontend.terminal.embed import InteractiveShellEmbed
    except ImportError:
        if debug:
            print_exc()
        return None
    return InteractiveShellEmbed.instance(banner1=banner, user_ns=overrides)


def start_ipython4(overrides, banner, *, debug=False):
    try:
        from IPython.terminal.embed import InteractiveShellEmbed
        shell = InteractiveShellEmbed.instance()
    except ImportError:
        if debug:
            print_exc()
        return None
    return partial(shell.mainloop, local_ns=overrides, display_banner=banner)


def start_ipython(overrides, banner, debug):
    shell = None
    for starter in (start_ipython4, start_ipython1):
        shell = starter(overrides, banner, debug=debug)
        if shell is not None:
            shell()
            break
    else:
        print(_('ipython is not available, set use_ipython to no'))


def start_python(overrides, banner):
    # Set the tab completion.
    with ExitStack() as resources:
        try:                                    # pragma: nocover
            import readline  # noqa: F401,E401
            import rlcompleter  # noqa: F401
        except ImportError:                     # pragma: nocover
            print(_('readline not available'), file=sys.stderr)
            pass
        else:
            readline.parse_and_bind('tab: complete')
            history_file_template = config.shell.history_file.strip()
            if len(history_file_template) > 0:
                # Expand substitutions.
                substitutions = {
                    key.lower(): value
                    for key, value in config.paths.items()
                    }
                history_file = Template(
                    history_file_template).safe_substitute(substitutions)
                with suppress(FileNotFoundError):
                    readline.read_history_file(history_file)
                resources.callback(
                    readline.write_history_file,
                    history_file)
        sys.ps1 = config.shell.prompt + ' '
        interact(upframe=False, banner=banner, overrides=overrides)


def do_interactive(ctx, banner):
    global m, r
    overrides = dict(
        m=m,
        commit=config.db.commit,
        abort=config.db.abort,
        config=config,
        getUtility=getUtility
        )
    # Bootstrap some useful names into the namespace, mostly to make
    # the component architecture and interfaces easily available.
    for module_name in sys.modules:
        if not module_name.startswith('mailman.interfaces.'):
            continue
        module = sys.modules[module_name]
        for name in module.__all__:
            overrides[name] = getattr(module, name)
    banner = config.shell.banner + '\n' + (
        banner if isinstance(banner, str) else '')
    try:
        use_ipython = as_boolean(config.shell.use_ipython)
    except ValueError:
        if config.shell.use_ipython == 'debug':
            use_ipython = True
            debug = True
        else:
            print(_('Invalid value for [shell]use_python: {}').format(
                config.shell.use_ipython), file=sys.stderr)
            return
    else:
        debug = False
    if use_ipython:
        start_ipython(overrides, banner, debug)
    else:
        start_python(overrides, banner)


def show_detailed_help(ctx, param, value):
    if not value:
        # Returning None tells click to process the rest of the command line.
        return
    # Split this up into paragraphs for easier translation.
    print(_("""\
This script provides you with a general framework for interacting with a
mailing list."""))
    print()
    print(_("""\
There are two ways to use this script: interactively or programmatically.
Using it interactively allows you to play with, examine and modify a mailing
list from Python's interactive interpreter.  When running interactively, the
variable 'm' will be available in the global namespace.  It will reference the
mailing list object."""))
    print()
    print(_("""\
Programmatically, you can write a function to operate on a mailing list, and
this script will take care of the housekeeping (see below for examples).  In
that case, the general usage syntax is:

% mailman withlist [options] -l listspec [args ...]

where `listspec` is either the posting address of the mailing list
(e.g. ant@example.com), or the List-ID (e.g. ant.example.com)."""))
    print()
    print(_("""\
Here's an example of how to use the --run option.  Say you have a file in the
Mailman installation directory called 'listaddr.py', with the following two
functions:

def listaddr(mlist):
    print(mlist.posting_address)

def requestaddr(mlist):
    print(mlist.request_address)

Run methods take at least one argument, the mailing list object to operate
on.  Any additional arguments given on the command line are passed as
positional arguments to the callable.

If -l is not given then you can run a function that takes no arguments.
"""))
    print()
    print(_("""\
You can print the list's posting address by running the following from the
command line:

% mailman withlist -r listaddr -l ant@example.com
Importing listaddr ...
Running listaddr.listaddr() ...
ant@example.com"""))
    print()
    print(_("""\
And you can print the list's request address by running:

% mailman withlist -r listaddr.requestaddr -l ant@example.com
Importing listaddr ...
Running listaddr.requestaddr() ...
ant-request@example.com"""))
    print()
    print(_("""\
As another example, say you wanted to change the display name for a particular
mailing list.  You could put the following function in a file called
`change.py`:

def change(mlist, display_name):
    mlist.display_name = display_name

and run this from the command line:

% mailman withlist -r change -l ant@example.com 'My List'

Note that you do not have to explicitly commit any database transactions, as
Mailman will do this for you (assuming no errors occured)."""))
    sys.exit(0)


@click.command(
    cls=I18nCommand,
    help=_("""\
    Operate on a mailing list.

    For detailed help, see --details
    """))
@click.option(
    '--interactive', '-i',
    is_flag=True, default=None,
    help=_("""\
    Leaves you at an interactive prompt after all other processing is complete.
    This is the default unless the --run option is given."""))
@click.option(
    '--run', '-r',
    help=_("""\

    Run a script.  The argument is the module path to a callable.  This
    callable will be imported and then, if --listspec/-l is also given, is
    called with the mailing list as the first argument.  If additional
    arguments are given at the end of the command line, they are passed as
    subsequent positional arguments to the callable.  For additional help, see
    --details.

    If no --listspec/-l argument is given, the script function being called is
    called with no arguments.
    """))
@click.option(
    '--details',
    is_flag=True, default=False, is_eager=True, expose_value=False,
    callback=show_detailed_help,
    help=_('Print detailed instructions and exit.'))
# Optional positional argument.
@click.option(
    '--listspec', '-l',
    help=_("""\
    A specification of the mailing list to operate on.  This may be the posting
    address of the list, or its List-ID.  The argument can also be a Python
    regular expression, in which case it is matched against both the posting
    address and List-ID of all mailing lists.  To use a regular expression,
    LISTSPEC must start with a ^ (and the matching is done with re.match().
    LISTSPEC cannot be a regular expression unless --run is given."""))
@click.argument('run_args', nargs=-1)
@click.pass_context
def shell(ctx, interactive, run, listspec, run_args):
    global m, r
    banner = DEFAULT_BANNER
    # Interactive is the default unless --run was given.
    interactive = (run is None) if interactive is None else interactive
    # List name cannot be a regular expression if --run is not given.
    if listspec and listspec.startswith('^') and not run:
        ctx.fail(_('Regular expression requires --run'))
    # Handle --run.
    list_manager = getUtility(IListManager)
    if run:
        # When the module and the callable have the same name, a shorthand
        # without the dot is allowed.
        dotted_name = (run if '.' in run else '{0}.{0}'.format(run))
        if listspec is None:
            r = call_name(dotted_name, *run_args)
        elif listspec.startswith('^'):
            r = {}
            cre = re.compile(listspec, re.IGNORECASE)
            for mlist in list_manager.mailing_lists:
                if cre.match(mlist.fqdn_listname) or cre.match(mlist.list_id):
                    results = call_name(dotted_name, mlist, *run_args)
                    r[mlist.list_id] = results
        else:
            m = list_manager.get(listspec)
            if m is None:
                ctx.fail(_('No such list: ${listspec}'))
            r = call_name(dotted_name, m, *run_args)
    else:
        # Not --run.
        if listspec is not None:
            m = list_manager.get(listspec)
            if m is None:
                ctx.fail(_('No such list: ${listspec}'))
            banner = _("The variable 'm' is the ${listspec} mailing list")
    # All other processing is finished; maybe go into interactive mode.
    if interactive:
        do_interactive(ctx, banner)


@public
@implementer(ICLISubCommand)
class Withlist:
    name = 'withlist'
    command = shell


@public
class Shell(Withlist):
    name = 'shell'
    command = shell
